# -*- coding: utf-8 -*-
# Copyright 2008-2018 pydicom authors. See LICENSE file for details.
"""Unit tests for the pydicom.dataelem module."""

# Many tests of DataElement class are implied in test_dataset also
import math

import pytest

from pydicom import filewriter, config, dcmread
from pydicom.charset import default_encoding
from pydicom.data import get_testdata_file
from pydicom.datadict import add_private_dict_entry
from pydicom.dataelem import (
    DataElement,
    RawDataElement,
    DataElement_from_raw,
)
from pydicom.dataset import Dataset
from pydicom.errors import BytesLengthException
from pydicom.filebase import DicomBytesIO
from pydicom.multival import MultiValue
from pydicom.tag import Tag, BaseTag
from pydicom.uid import UID
from pydicom.valuerep import DSfloat


class TestDataElement:
    """Tests for dataelem.DataElement."""
    def setup(self):
        self.data_elementSH = DataElement((1, 2), "SH", "hello")
        self.data_elementIS = DataElement((1, 2), "IS", "42")
        self.data_elementDS = DataElement((1, 2), "DS", "42.00001")
        self.data_elementMulti = DataElement((1, 2), "DS",
                                             ['42.1', '42.2', '42.3'])
        self.data_elementCommand = DataElement(0x00000000, 'UL', 100)
        self.data_elementPrivate = DataElement(0x00090000, 'UL', 101)
        self.data_elementRetired = DataElement(0x00080010, 'SH', 102)
        config.use_none_as_empty_text_VR_value = False

    def teardown(self):
        config.use_none_as_empty_text_VR_value = False

    @pytest.fixture
    def replace_un_with_known_vr(self):
        old_value = config.replace_un_with_known_vr
        config.replace_un_with_known_vr = True
        yield
        config.replace_un_with_known_vr = old_value

    def test_AT(self):
        """VR of AT takes Tag variants when set"""
        elem1 = DataElement("OffendingElement", "AT", 0x100010)
        elem2 = DataElement("OffendingElement", "AT", (0x10, 0x10))
        elem3 = DataElement(
            "FrameIncrementPointer", "AT", [0x540010, 0x540020]
        )
        elem4 = DataElement("OffendingElement", "AT", "PatientName")
        assert isinstance(elem1.value, BaseTag)
        assert isinstance(elem2.value, BaseTag)
        assert elem1.value == elem2.value == elem4.value
        assert elem1.value == 0x100010
        assert isinstance(elem3.value, MultiValue)
        assert len(elem3.value) == 2

        # Test also using Dataset, and check 0x00000000 works
        ds = Dataset()
        ds.OffendingElement = 0
        assert isinstance(ds.OffendingElement, BaseTag)
        ds.OffendingElement = (0x0000, 0x0000)
        assert isinstance(ds.OffendingElement, BaseTag)
        assert ds.OffendingElement == 0

        # An invalid Tag should throw an error
        with pytest.raises(OverflowError):
            _ = DataElement("OffendingElement", "AT", 0x100000000)

    def test_VM_1(self):
        """DataElement: return correct value multiplicity for VM > 1"""
        assert 3 == self.data_elementMulti.VM

    def test_VM_2(self):
        """DataElement: return correct value multiplicity for VM = 1"""
        assert 1 == self.data_elementIS.VM

    def test_DSFloat_conversion(self):
        """Test that strings are correctly converted if changing the value."""
        assert isinstance(self.data_elementDS.value, DSfloat)
        assert isinstance(self.data_elementMulti.value[0], DSfloat)
        assert DSfloat('42.1') == self.data_elementMulti.value[0]

        # multi-value append/insert
        self.data_elementMulti.value.append('42.4')
        assert isinstance(self.data_elementMulti.value[3], DSfloat)
        assert DSfloat('42.4') == self.data_elementMulti.value[3]

        self.data_elementMulti.value.insert(0, '42.0')
        assert isinstance(self.data_elementMulti.value[0], DSfloat)
        assert DSfloat('42.0') == self.data_elementMulti.value[0]

        # change single value of multi-value
        self.data_elementMulti.value[3] = '123.4'
        assert isinstance(self.data_elementMulti.value[3], DSfloat)
        assert DSfloat('123.4') == self.data_elementMulti.value[3]

    def test_DSFloat_conversion_auto_format(self):
        """Test that strings are being auto-formatted correctly."""
        data_element = DataElement((1, 2), "DS",
                                   DSfloat(math.pi, auto_format=True))
        assert math.pi == data_element.value
        assert '3.14159265358979' == str(data_element.value)

    def test_backslash(self):
        """DataElement: String with '\\' sets multi-valued data_element."""
        data_element = DataElement((1, 2), "DS", r"42.1\42.2\42.3")
        assert 3 == data_element.VM

    def test_UID(self):
        """DataElement: setting or changing UID results in UID type."""
        ds = Dataset()
        ds.TransferSyntaxUID = "1.2.3"
        assert isinstance(ds.TransferSyntaxUID, UID)
        ds.TransferSyntaxUID += ".4.5.6"
        assert isinstance(ds.TransferSyntaxUID, UID)

    def test_keyword(self):
        """DataElement: return correct keyword"""
        assert 'CommandGroupLength' == self.data_elementCommand.keyword
        assert '' == self.data_elementPrivate.keyword

    def test_retired(self):
        """DataElement: return correct is_retired"""
        assert self.data_elementCommand.is_retired is False
        assert self.data_elementRetired.is_retired is True
        assert self.data_elementPrivate.is_retired is False

    def test_description_group_length(self):
        """Test DataElement.description for Group Length element"""
        elem = DataElement(0x00100000, 'LO', 12345)
        assert 'Group Length' == elem.description()

    def test_description_unknown_private(self):
        """Test DataElement.description with an unknown private element"""
        elem = DataElement(0x00110010, 'LO', 12345)
        elem.private_creator = 'TEST'
        assert 'Private tag data' == elem.description()
        elem = DataElement(0x00110F00, 'LO', 12345)
        assert elem.tag.is_private
        assert elem.private_creator is None
        assert 'Private tag data' == elem.description()

    def test_description_unknown(self):
        """Test DataElement.description with an unknown element"""
        elem = DataElement(0x00000004, 'LO', 12345)
        assert '' == elem.description()

    def test_equality_standard_element(self):
        """DataElement: equality returns correct value for simple elements"""
        dd = DataElement(0x00100010, 'PN', 'ANON')
        assert dd == dd
        ee = DataElement(0x00100010, 'PN', 'ANON')
        assert dd == ee

        # Check value
        ee.value = 'ANAN'
        assert not dd == ee

        # Check tag
        ee = DataElement(0x00100011, 'PN', 'ANON')
        assert not dd == ee

        # Check VR
        ee = DataElement(0x00100010, 'SH', 'ANON')
        assert not dd == ee

        dd = DataElement(0x00080018, 'UI', '1.2.3.4')
        ee = DataElement(0x00080018, 'UI', '1.2.3.4')
        assert dd == ee

        ee = DataElement(0x00080018, 'PN', '1.2.3.4')
        assert not dd == ee

    def test_equality_private_element(self):
        """DataElement: equality returns correct value for private elements"""
        dd = DataElement(0x01110001, 'PN', 'ANON')
        assert dd == dd
        ee = DataElement(0x01110001, 'PN', 'ANON')
        assert dd == ee

        # Check value
        ee.value = 'ANAN'
        assert not dd == ee

        # Check tag
        ee = DataElement(0x01110002, 'PN', 'ANON')
        assert not dd == ee

        # Check VR
        ee = DataElement(0x01110001, 'SH', 'ANON')
        assert not dd == ee

    def test_equality_sequence_element(self):
        """DataElement: equality returns correct value for sequence elements"""
        dd = DataElement(0x300A00B0, 'SQ', [])
        assert dd == dd
        ee = DataElement(0x300A00B0, 'SQ', [])
        assert dd == ee

        # Check value
        e = Dataset()
        e.PatientName = 'ANON'
        ee.value = [e]
        assert not dd == ee

        # Check tag
        ee = DataElement(0x01110002, 'SQ', [])
        assert not dd == ee

        # Check VR
        ee = DataElement(0x300A00B0, 'SH', [])
        assert not dd == ee

        # Check with dataset
        dd = DataElement(0x300A00B0, 'SQ', [Dataset()])
        dd.value[0].PatientName = 'ANON'
        ee = DataElement(0x300A00B0, 'SQ', [Dataset()])
        ee.value[0].PatientName = 'ANON'
        assert dd == ee

        # Check uneven sequences
        dd.value.append(Dataset())
        dd.value[1].PatientName = 'ANON'
        assert not dd == ee

        ee.value.append(Dataset())
        ee.value[1].PatientName = 'ANON'
        assert dd == ee
        ee.value.append(Dataset())
        ee.value[2].PatientName = 'ANON'
        assert not dd == ee

    def test_equality_not_rlement(self):
        """DataElement: equality returns correct value when not same class"""
        dd = DataElement(0x00100010, 'PN', 'ANON')
        ee = {'0x00100010': 'ANON'}
        assert not dd == ee

    def test_equality_inheritance(self):
        """DataElement: equality returns correct value for subclasses"""

        class DataElementPlus(DataElement):
            pass

        dd = DataElement(0x00100010, 'PN', 'ANON')
        ee = DataElementPlus(0x00100010, 'PN', 'ANON')
        assert ee == ee
        assert dd == ee
        assert ee == dd

        ee = DataElementPlus(0x00100010, 'PN', 'ANONY')
        assert not dd == ee
        assert not ee == dd

    def test_equality_class_members(self):
        """Test equality is correct when ignored class members differ."""
        dd = DataElement(0x00100010, 'PN', 'ANON')
        dd.showVR = False
        dd.file_tell = 10
        dd.maxBytesToDisplay = 0
        dd.descripWidth = 0
        assert DataElement(0x00100010, 'PN', 'ANON') == dd

    def test_inequality_standard(self):
        """Test DataElement.__ne__ for standard element"""
        dd = DataElement(0x00100010, 'PN', 'ANON')
        assert not dd != dd
        assert DataElement(0x00100010, 'PN', 'ANONA') != dd

        # Check tag
        assert DataElement(0x00100011, 'PN', 'ANON') != dd

        # Check VR
        assert DataElement(0x00100010, 'SH', 'ANON') != dd

    def test_inequality_sequence(self):
        """Test DataElement.__ne__ for sequence element"""
        dd = DataElement(0x300A00B0, 'SQ', [])
        assert not dd != dd
        assert not DataElement(0x300A00B0, 'SQ', []) != dd
        ee = DataElement(0x300A00B0, 'SQ', [Dataset()])
        assert ee != dd

        # Check value
        dd.value = [Dataset()]
        dd[0].PatientName = 'ANON'
        ee[0].PatientName = 'ANON'
        assert not ee != dd
        ee[0].PatientName = 'ANONA'
        assert ee != dd

    def test_hash(self):
        """Test hash(DataElement) raises TypeError"""
        with pytest.raises(TypeError, match=r"unhashable"):
            hash(DataElement(0x00100010, 'PN', 'ANON'))

    def test_repeater_str(self):
        """Test a repeater group element displays the element name."""
        elem = DataElement(0x60023000, 'OB', b'\x00')
        assert 'Overlay Data' in elem.__str__()

    def test_str_no_vr(self):
        """Test DataElement.__str__ output with no VR"""
        elem = DataElement(0x00100010, 'PN', 'ANON')
        assert "(0010, 0010) Patient's Name" in str(elem)
        assert "PN: 'ANON'" in str(elem)
        elem.showVR = False
        assert "(0010, 0010) Patient's Name" in str(elem)
        assert 'PN' not in str(elem)

    def test_repr_seq(self):
        """Test DataElement.__repr__ with a sequence"""
        elem = DataElement(0x300A00B0, 'SQ', [Dataset()])
        elem[0].PatientID = '1234'
        assert repr(elem) == repr(elem.value)

    def test_getitem_raises(self):
        """Test DataElement.__getitem__ raise if value not indexable"""
        elem = DataElement(0x00100010, 'LO', 12345)
        with pytest.raises(TypeError):
            elem[0]

    def test_repval_large_elem(self):
        """Test DataElement.repval doesn't return a huge string for a large
        value"""
        elem = DataElement(0x00820003, 'UT', 'a' * 1000)
        assert len(elem.repval) < 100

    def test_repval_large_vm(self):
        """Test DataElement.repval doesn't return a huge string for a large
        vm"""
        elem = DataElement(0x00080054, 'AE', 'a\\' * 1000 + 'a')
        assert len(elem.repval) < 100

    def test_repval_strange_type(self):
        """Test DataElement.repval doesn't break with bad types"""
        elem = DataElement(0x00020001, 'OB', 0)
        assert len(elem.repval) < 100

    def test_private_tag_in_repeater_range(self):
        """Test that an unknown private tag (e.g. a tag not in the private
        dictionary) in the repeater range is not handled as a repeater tag
        if using Implicit Little Endian transfer syntax."""
        # regression test for #689
        ds = Dataset()
        ds[0x50f10010] = RawDataElement(
            Tag(0x50f10010), None, 8, b'FDMS 1.0', 0, True, True)
        ds[0x50f1100a] = RawDataElement(
            Tag(0x50f1100a), None, 6, b'ACC0.6', 0, True, True)
        private_creator_data_elem = ds[0x50f10010]
        assert 'Private Creator' == private_creator_data_elem.name
        assert 'LO' == private_creator_data_elem.VR

        private_data_elem = ds[0x50f1100a]
        assert '[FNC Parameters]' == private_data_elem.name
        assert 'SH' == private_data_elem.VR

    def test_private_repeater_tag(self):
        """Test that a known private tag in the repeater range is correctly
        handled using Implicit Little Endian transfer syntax."""
        ds = Dataset()
        ds[0x60210012] = RawDataElement(
            Tag(0x60210012), None, 12, b'PAPYRUS 3.0 ', 0, True, True)
        ds[0x60211200] = RawDataElement(
            Tag(0x60211200), None, 6, b'123456', 0, True, True)
        private_creator_data_elem = ds[0x60210012]
        assert 'Private Creator' == private_creator_data_elem.name
        assert 'LO' == private_creator_data_elem.VR

        private_data_elem = ds[0x60211200]
        assert '[Overlay ID]' == private_data_elem.name
        assert 'IS' == private_data_elem.VR

    def test_known_tags_with_UN_VR(self, replace_un_with_known_vr):
        """Known tags with VR UN are correctly decoded."""
        ds = Dataset()
        ds[0x00080005] = DataElement(0x00080005, 'UN', b'ISO_IR 126')
        ds[0x00100010] = DataElement(0x00100010, 'UN',
                                     'Διονυσιος'.encode('iso_ir_126'))
        ds.decode()
        assert 'CS' == ds[0x00080005].VR
        assert 'PN' == ds[0x00100010].VR
        assert 'Διονυσιος' == ds[0x00100010].value

        ds = Dataset()
        ds[0x00080005] = DataElement(0x00080005, 'UN',
                                     b'ISO 2022 IR 100\\ISO 2022 IR 126')
        ds[0x00100010] = DataElement(0x00100010, 'UN',
                                     b'Dionysios=\x1b\x2d\x46'
                                     + 'Διονυσιος'.encode('iso_ir_126'))
        ds.decode()
        assert 'CS' == ds[0x00080005].VR
        assert 'PN' == ds[0x00100010].VR
        assert 'Dionysios=Διονυσιος' == ds[0x00100010].value

    def test_reading_ds_with_known_tags_with_UN_VR(
            self, replace_un_with_known_vr):
        """Known tags with VR UN are correctly read."""
        test_file = get_testdata_file('explicit_VR-UN.dcm')
        ds = dcmread(test_file)
        assert 'CS' == ds[0x00080005].VR
        assert 'TM' == ds[0x00080030].VR
        assert 'PN' == ds[0x00100010].VR
        assert 'PN' == ds[0x00100010].VR
        assert 'DA' == ds[0x00100030].VR

    def test_unknown_tags_with_UN_VR(self):
        """Unknown tags with VR UN are not decoded."""
        ds = Dataset()
        ds[0x00080005] = DataElement(0x00080005, 'CS', b'ISO_IR 126')
        ds[0x00111010] = DataElement(0x00111010, 'UN',
                                     'Διονυσιος'.encode('iso_ir_126'))
        ds.decode()
        assert 'UN' == ds[0x00111010].VR
        assert 'Διονυσιος'.encode('iso_ir_126') == ds[0x00111010].value

    def test_tag_with_long_value_UN_VR(self):
        """Tag with length > 64kb with VR UN is not changed."""
        ds = Dataset()
        ds[0x00080005] = DataElement(0x00080005, 'CS', b'ISO_IR 126')

        single_value = b'123456.789012345'
        large_value = b'\\'.join([single_value] * 4500)
        ds[0x30040058] = DataElement(0x30040058, 'UN',
                                     large_value,
                                     is_undefined_length=False)
        ds.decode()
        assert 'UN' == ds[0x30040058].VR

    @pytest.mark.parametrize('use_none, empty_value',
                             ((True, None), (False, '')))
    def test_empty_text_values(self, use_none, empty_value,
                               no_datetime_conversion):
        """Test that assigning an empty value behaves as expected."""
        def check_empty_text_element(value):
            setattr(ds, tag_name, value)
            elem = ds[tag_name]
            assert bool(elem.value) is False
            assert 0 == elem.VM
            assert elem.value == value
            fp = DicomBytesIO()
            filewriter.write_dataset(fp, ds)
            ds_read = dcmread(fp, force=True)
            assert empty_value == ds_read[tag_name].value

        text_vrs = {
            'AE': 'RetrieveAETitle',
            'AS': 'PatientAge',
            'CS': 'QualityControlSubject',
            'DA': 'PatientBirthDate',
            'DT': 'AcquisitionDateTime',
            'LO': 'DataSetSubtype',
            'LT': 'ExtendedCodeMeaning',
            'PN': 'PatientName',
            'SH': 'CodeValue',
            'ST': 'InstitutionAddress',
            'TM': 'StudyTime',
            'UC': 'LongCodeValue',
            'UI': 'SOPClassUID',
            'UR': 'CodingSchemeURL',
            'UT': 'StrainAdditionalInformation',
        }
        config.use_none_as_empty_text_VR_value = use_none
        ds = Dataset()
        ds.is_little_endian = True
        # set value to new element
        for tag_name in text_vrs.values():
            check_empty_text_element(None)
            del ds[tag_name]
            check_empty_text_element('')
            del ds[tag_name]
            check_empty_text_element([])
            del ds[tag_name]

        # set value to existing element
        for tag_name in text_vrs.values():
            check_empty_text_element(None)
            check_empty_text_element('')
            check_empty_text_element([])
            check_empty_text_element(None)

    def test_empty_binary_values(self):
        """Test that assigning an empty value behaves as expected for
        non-text VRs."""
        def check_empty_binary_element(value):
            setattr(ds, tag_name, value)
            elem = ds[tag_name]
            assert bool(elem.value) is False
            assert 0 == elem.VM
            assert elem.value == value
            fp = DicomBytesIO()
            filewriter.write_dataset(fp, ds)
            ds_read = dcmread(fp, force=True)
            assert ds_read[tag_name].value is None

        non_text_vrs = {
            'AT': 'OffendingElement',
            'DS': 'PatientWeight',
            'IS': 'BeamNumber',
            'SL': 'RationalNumeratorValue',
            'SS': 'SelectorSSValue',
            'UL': 'SimpleFrameList',
            'US': 'SourceAcquisitionBeamNumber',
            'FD': 'RealWorldValueLUTData',
            'FL': 'VectorAccuracy',
            'OB': 'FillPattern',
            'OD': 'DoubleFloatPixelData',
            'OF': 'UValueData',
            'OL': 'TrackPointIndexList',
            'OW': 'TrianglePointIndexList',
            'UN': 'SelectorUNValue',
        }
        ds = Dataset()
        ds.is_little_endian = True
        # set value to new element
        for tag_name in non_text_vrs.values():
            check_empty_binary_element(None)
            del ds[tag_name]
            check_empty_binary_element([])
            del ds[tag_name]
            check_empty_binary_element(MultiValue(int, []))
            del ds[tag_name]

        # set value to existing element
        for tag_name in non_text_vrs.values():
            check_empty_binary_element(None)
            check_empty_binary_element([])
            check_empty_binary_element(MultiValue(int, []))
            check_empty_binary_element(None)

    def test_empty_sequence_is_handled_as_array(self):
        ds = Dataset()
        ds.AcquisitionContextSequence = []
        elem = ds['AcquisitionContextSequence']
        assert bool(elem.value) is False
        assert 0 == elem.VM
        assert elem.value == []

        fp = DicomBytesIO()
        fp.is_little_endian = True
        fp.is_implicit_VR = True
        filewriter.write_dataset(fp, ds)
        ds_read = dcmread(fp, force=True)
        elem = ds_read['AcquisitionContextSequence']
        assert 0 == elem.VM
        assert elem.value == []

    def test_is_private(self):
        """Test the is_private property."""
        elem = DataElement(0x00090010, 'UN', None)
        assert elem.is_private
        elem = DataElement(0x00080010, 'UN', None)
        assert not elem.is_private


class TestRawDataElement:

    """Tests for dataelem.RawDataElement."""
    def test_invalid_tag_warning(self, allow_invalid_values):
        """RawDataElement: conversion of unknown tag warns..."""
        raw = RawDataElement(Tag(0x88880088), None, 4, b'unknown',
                             0, True, True)

        with pytest.warns(UserWarning, match=r"\(8888, 0088\)"):
            element = DataElement_from_raw(raw)
            assert element.VR == 'UN'

    def test_key_error(self, enforce_valid_values):
        """RawDataElement: conversion of unknown tag throws KeyError..."""
        # raw data element -> tag VR length value
        #                       value_tell is_implicit_VR is_little_endian'
        # Unknown (not in DICOM dict), non-private, non-group 0 for this test
        raw = RawDataElement(Tag(0x88880002), None, 4, b'unknown',
                             0, True, True)

        with pytest.raises(KeyError, match=r"\(8888, 0002\)"):
            DataElement_from_raw(raw)

    def test_valid_tag(self, no_datetime_conversion):
        """RawDataElement: conversion of known tag succeeds..."""
        raw = RawDataElement(Tag(0x00080020), 'DA', 8, b'20170101',
                             0, False, True)
        element = DataElement_from_raw(raw, default_encoding)
        assert 'Study Date' == element.name
        assert 'DA' == element.VR
        assert '20170101' == element.value

        raw = RawDataElement(Tag(0x00080000), None, 4, b'\x02\x00\x00\x00',
                             0, True, True)
        elem = DataElement_from_raw(raw, default_encoding)
        assert 'UL' == elem.VR

    def test_data_element_without_encoding(self):
        """RawDataElement: no encoding needed."""
        raw = RawDataElement(Tag(0x00104000), 'LT', 23,
                             b'comment\\comment2\\comment3',
                             0, False, True)
        element = DataElement_from_raw(raw)
        assert 'Patient Comments' == element.name

    def test_unknown_vr(self):
        """Test converting a raw element with unknown VR"""
        raw = RawDataElement(Tag(0x00080000), 'AA', 8, b'20170101',
                             0, False, True)
        with pytest.raises(NotImplementedError):
            DataElement_from_raw(raw, default_encoding)

    @pytest.fixture
    def accept_wrong_length(self, request):
        old_value = config.convert_wrong_length_to_UN
        config.convert_wrong_length_to_UN = request.param
        yield
        config.convert_wrong_length_to_UN = old_value

    @pytest.mark.parametrize("accept_wrong_length", [False], indirect=True)
    def test_wrong_bytes_length_exception(self, accept_wrong_length):
        """Check exception when number of raw bytes is not correct."""
        raw = RawDataElement(Tag(0x00190000), 'FD', 1, b'1', 0, False, True)
        with pytest.raises(BytesLengthException):
            DataElement_from_raw(raw)

    @pytest.mark.parametrize("accept_wrong_length", [True], indirect=True)
    def test_wrong_bytes_length_convert_to_UN(self, accept_wrong_length):
        """Check warning and behavior for incorrect number of raw bytes."""
        value = b'1'
        raw = RawDataElement(Tag(0x00190000), 'FD', 1, value, 0, False, True)
        msg = (
            r"Expected total bytes to be an even multiple of bytes per value. "
            r"Instead received b'1' with length 1 and struct format 'd' which "
            r"corresponds to bytes per value of 8. This occurred while trying "
            r"to parse \(0019, 0000\) according to VR 'FD'. "
            r"Setting VR to 'UN'."
        )
        with pytest.warns(UserWarning, match=msg):
            raw_elem = DataElement_from_raw(raw)
            assert 'UN' == raw_elem.VR
            assert value == raw_elem.value

    def test_read_known_private_tag_implicit(self):
        fp = DicomBytesIO()
        ds = Dataset()
        ds.is_implicit_VR = True
        ds.is_little_endian = True
        ds[0x00410010] = RawDataElement(
            Tag(0x00410010), "LO", 8, b"ACME 3.2", 0, True, True)
        ds[0x00411001] = RawDataElement(
            Tag(0x00411001), "US", 2, b"\x2A\x00", 0, True, True)
        ds[0x00431001] = RawDataElement(
            Tag(0x00431001), "SH", 8, b"Unknown ", 0, True, True)
        ds.save_as(fp)
        ds = dcmread(fp, force=True)
        elem = ds[0x00411001]
        assert elem.VR == "UN"
        assert elem.name == "Private tag data"
        assert elem.value == b"\x2A\x00"

        add_private_dict_entry("ACME 3.2", 0x00410001, "US", "Some Number")
        ds = dcmread(fp, force=True)
        elem = ds[0x00411001]
        assert elem.VR == "US"
        assert elem.name == "[Some Number]"
        assert elem.value == 42

        # Unknown private tag is handled as before
        elem = ds[0x00431001]
        assert elem.VR == "UN"
        assert elem.name == "Private tag data"
        assert elem.value == b"Unknown "

    def test_read_known_private_tag_explicit(self):
        fp = DicomBytesIO()
        ds = Dataset()
        ds.is_implicit_VR = False
        ds.is_little_endian = True
        ds[0x00410010] = RawDataElement(
            Tag(0x00410010), "LO", 8, b"ACME 3.2", 0, False, True)
        ds[0x00411002] = RawDataElement(
            Tag(0x00411002), "UN", 8, b"SOME_AET", 0, False, True)
        ds.save_as(fp)
        ds = dcmread(fp, force=True)
        elem = ds[0x00411002]
        assert elem.VR == "UN"
        assert elem.name == "Private tag data"
        assert elem.value == b"SOME_AET"

        add_private_dict_entry("ACME 3.2", 0x00410002, "AE", "Some AET")
        ds = dcmread(fp, force=True)
        elem = ds[0x00411002]
        assert elem.VR == "AE"
        assert elem.name == "[Some AET]"
        assert elem.value == "SOME_AET"

    def test_read_known_private_tag_explicit_no_lookup(
            self, dont_replace_un_with_known_vr):
        add_private_dict_entry("ACME 3.2", 0x00410003, "IS", "Another Number")
        fp = DicomBytesIO()
        ds = Dataset()
        ds.is_implicit_VR = False
        ds.is_little_endian = True
        ds[0x00410010] = RawDataElement(
            Tag(0x00410010), "LO", 8, b"ACME 3.2", 0, False, True)
        ds[0x00411003] = RawDataElement(
            Tag(0x00411003), "UN", 8, b"12345678", 0, False, True)
        ds.save_as(fp)
        ds = dcmread(fp, force=True)
        elem = ds[0x00411003]
        assert elem.VR == "UN"
        assert elem.name == "[Another Number]"
        assert elem.value == b"12345678"



# Copyright 2008-2020 pydicom authors. See LICENSE file for details.
"""Code for multi-value data elements values,
or any list of items that must all be the same type.
"""

from typing import (
    Iterable, Union, List, overload, Callable, Any, cast,
    TypeVar, MutableSequence, Iterator
)
from typing import Sequence as SequenceType


_T = TypeVar("_T")
_ItemType = TypeVar("_ItemType")


class MultiValue(MutableSequence[_ItemType]):
    """Class to hold any multi-valued DICOM value, or any list of items that
    are all of the same type.

    This class enforces that any items added to the list are of the correct
    type, by calling the constructor on any items that are added. Therefore,
    the constructor must behave nicely if passed an object that is already its
    type. The constructor should raise :class:`TypeError` if the item cannot be
    converted.

    Note, however, that DS and IS types can be a blank string ``''`` rather
    than an instance of their classes.
    """

    def __init__(
        self,
        type_constructor: Callable[[_T], _ItemType],
        iterable: Iterable[_T]
    ) -> None:
        """Create a new :class:`MultiValue` from an iterable and ensure each
        item in the :class:`MultiValue` has the same type.

        Parameters
        ----------
        type_constructor : callable
            A constructor for the required type for all items. Could be
            the class, or a factory function. Takes a single parameter and
            returns the input as the desired type (or raises an appropriate
            exception).
        iterable : iterable
            An iterable (e.g. :class:`list`, :class:`tuple`) of items to
            initialize the :class:`MultiValue` list. Each item in the iterable
            is passed to `type_constructor` and the returned value added to
            the :class:`MultiValue`.
        """
        from pydicom.valuerep import DSfloat, DSdecimal, IS

        def DS_IS_constructor(x: _T) -> _ItemType:
            return self.type_constructor(x) if x != '' else cast(_ItemType, x)

        self._list: List[_ItemType] = list()
        self.type_constructor = type_constructor
        if type_constructor in (DSfloat, IS, DSdecimal):
            type_constructor = DS_IS_constructor

        for x in iterable:
            self._list.append(type_constructor(x))

    def append(self, val: _T) -> None:
        self._list.append(self.type_constructor(val))

    def __delitem__(self, index: Union[slice, int]) -> None:
        del self._list[index]

    def extend(self, val: Iterable[_T]) -> None:
        """Extend the :class:`~pydicom.multival.MultiValue` using an iterable
        of objects.
        """
        self._list.extend([self.type_constructor(x) for x in val])

    def __iadd__(self, other: Iterable[_T]) -> MutableSequence[_ItemType]:
        """Implement MultiValue() += Iterable[Any]."""
        self._list += [self.type_constructor(x) for x in other]
        return self

    def __eq__(self, other: Any) -> Any:
        return self._list == other

    @overload
    def __getitem__(self, index: int) -> _ItemType: pass  # pragma: no cover

    @overload
    def __getitem__(self, index: slice) -> MutableSequence[_ItemType]:
        pass  # pragma: no cover

    def __getitem__(
        self, index: Union[slice, int]
    ) -> Union[MutableSequence[_ItemType], _ItemType]:
        return self._list[index]

    def insert(self, position: int, val: _T) -> None:
        self._list.insert(position, self.type_constructor(val))

    def __iter__(self) -> Iterator[_ItemType]:
        yield from self._list

    def __len__(self) -> int:
        return len(self._list)

    def __ne__(self, other: Any) -> Any:
        return self._list != other

    @overload
    def __setitem__(self, idx: int, val: _T) -> None: pass  # pragma: no cover

    @overload
    def __setitem__(self, idx: slice, val: Iterable[_T]) -> None:
        pass  # pragma: no cover

    def __setitem__(  # type: ignore[misc]
        self, idx: Union[int, slice], val: Union[_T, Iterable[_T]]
    ) -> None:
        """Set an item of the list, making sure it is of the right VR type"""
        if isinstance(idx, slice):
            val = cast(Iterable[_T], val)
            out = [self.type_constructor(v) for v in val]
            self._list.__setitem__(idx, out)
        else:
            val = cast(_T, val)
            self._list.__setitem__(idx, self.type_constructor(val))

    def sort(self, *args: Any, **kwargs: Any) -> None:
        self._list.sort(*args, **kwargs)

    def __str__(self) -> str:
        if not self:
            return ''
        lines = (
            f"{x!r}" if isinstance(x, (str, bytes)) else str(x) for x in self
        )
        return f"[{', '.join(lines)}]"

    __repr__ = __str__



# Copyright 2008-2018 pydicom authors. See LICENSE file for details.
"""Encoding benchmarks for the rle_handler module."""

from pydicom import dcmread
from pydicom.data import get_testdata_file
from pydicom.pixel_data_handlers.rle_handler import rle_encode_frame
from pydicom.uid import RLELossless


# 8/8-bit, 1 sample/pixel, 1 frame
EXPL_8_1_1F = get_testdata_file("OBXXXX1A.dcm")
# 8/8-bit, 3 sample/pixel, 1 frame
EXPL_8_3_1F = get_testdata_file("SC_rgb.dcm")
# 16/16-bit, 1 sample/pixel, 1 frame
EXPL_16_1_1F = get_testdata_file("MR_small.dcm")
# 16/16-bit, 3 sample/pixel, 1 frame
EXPL_16_3_1F = get_testdata_file("SC_rgb_16bit.dcm")
# 32/32-bit, 1 sample/pixel, 1 frame
EXPL_32_1_1F = get_testdata_file("rtdose_1frame.dcm")
# 32/32-bit, 3 sample/pixel, 1 frame
EXPL_32_3_1F = get_testdata_file("SC_rgb_32bit.dcm")


class TimeRLEEncodeFrame:
    """Time tests for rle_handler.rle_encode_frame."""
    def setup(self):
        ds = dcmread(EXPL_8_1_1F)
        self.arr8_1 = ds.pixel_array
        ds = dcmread(EXPL_8_3_1F)
        self.arr8_3 = ds.pixel_array
        ds = dcmread(EXPL_16_1_1F)
        self.arr16_1 = ds.pixel_array
        ds = dcmread(EXPL_16_3_1F)
        self.arr16_3 = ds.pixel_array
        ds = dcmread(EXPL_32_1_1F)
        self.arr32_1 = ds.pixel_array
        ds = dcmread(EXPL_32_3_1F)
        self.arr32_3 = ds.pixel_array

        self.no_runs = 100

    def time_08_1(self):
        """Time encoding 8 bit 1 sample/pixel."""
        for ii in range(self.no_runs):
            rle_encode_frame(self.arr8_1)

    def time_08_3(self):
        """Time encoding 8 bit 3 sample/pixel."""
        for ii in range(self.no_runs):
            rle_encode_frame(self.arr8_3)

    def time_16_1(self):
        """Time encoding 16 bit 1 sample/pixel."""
        for ii in range(self.no_runs):
            rle_encode_frame(self.arr16_1)

    def time_16_3(self):
        """Time encoding 16 bit 3 sample/pixel."""
        for ii in range(self.no_runs):
            rle_encode_frame(self.arr16_3)

    def time_32_1(self):
        """Time encoding 32 bit 1 sample/pixel."""
        for ii in range(self.no_runs):
            rle_encode_frame(self.arr32_1)

    def time_32_3(self):
        """Time encoding 32 bit 3 sample/pixel."""
        for ii in range(self.no_runs):
            rle_encode_frame(self.arr32_3)


# Requires numpy, pylibjpeg, pylibjpeg-rle and python-gdcm
class TimeDatasetCompress:
    """Test Dataset.compress()."""
    def setup(self):
        # More real-world like dataset
        self.ds = dcmread(EXPL_8_1_1F)
        self.arr8_1 = self.ds.pixel_array

        self.no_runs = 1000

    def time_pydicom(self):
        """Time the native RLE encoder."""
        for _ in range(self.no_runs):
            self.ds.compress(
                RLELossless, self.arr8_1, encoding_plugin='pydicom'
            )

    def time_pylibjpeg(self):
        """Time the pylibjpeg-rle Rust RLE encoder."""
        for _ in range(self.no_runs):
            self.ds.compress(
                RLELossless, self.arr8_1, encoding_plugin='pylibjpeg'
            )

    def time_gdcm(self):
        """Time the GDCM C++ RLE encoder."""
        for _ in range(self.no_runs):
            self.ds.compress(
                RLELossless, self.arr8_1, encoding_plugin='gdcm'
            )



# Copyright 2008-2018 pydicom authors. See LICENSE file for details.
"""Use the `GDCM <http://gdcm.sourceforge.net/>`_ Python package to decode
pixel transfer syntaxes.
"""

from tempfile import NamedTemporaryFile
from typing import TYPE_CHECKING, cast

if TYPE_CHECKING:  # pragma: no cover
    from pydicom.dataset import Dataset, FileMetaDataset, FileDataset


try:
    import numpy
    HAVE_NP = True
except ImportError:
    HAVE_NP = False

try:
    import gdcm
    from gdcm import DataElement
    HAVE_GDCM = True
    HAVE_GDCM_IN_MEMORY_SUPPORT = hasattr(DataElement, 'SetByteStringValue')
except ImportError:
    HAVE_GDCM = False
    HAVE_GDCM_IN_MEMORY_SUPPORT = False

from pydicom import config
from pydicom.encaps import generate_pixel_data
import pydicom.uid
from pydicom.uid import UID, JPEG2000, JPEG2000Lossless
from pydicom.pixel_data_handlers.util import (
    get_expected_length, pixel_dtype, get_j2k_parameters
)


HANDLER_NAME = 'GDCM'

DEPENDENCIES = {
    'numpy': ('http://www.numpy.org/', 'NumPy'),
    'gdcm': ('http://gdcm.sourceforge.net/wiki/index.php/Main_Page', 'GDCM'),
}

SUPPORTED_TRANSFER_SYNTAXES = [
    pydicom.uid.JPEGBaseline8Bit,
    pydicom.uid.JPEGExtended12Bit,
    pydicom.uid.JPEGLosslessP14,
    pydicom.uid.JPEGLosslessSV1,
    pydicom.uid.JPEGLSLossless,
    pydicom.uid.JPEGLSNearLossless,
    pydicom.uid.JPEG2000Lossless,
    pydicom.uid.JPEG2000,
]

should_convert_these_syntaxes_to_RGB = [pydicom.uid.JPEGBaseline8Bit]


def is_available() -> bool:
    """Return ``True`` if the handler has its dependencies met."""
    return HAVE_NP and HAVE_GDCM


def needs_to_convert_to_RGB(ds: "Dataset") -> bool:
    """Return ``True`` if the *Pixel Data* should to be converted from YCbCr to
    RGB.

    This affects JPEG transfer syntaxes.
    """
    return False


def should_change_PhotometricInterpretation_to_RGB(ds: "Dataset") -> bool:
    """Return ``True`` if the *Photometric Interpretation* should be changed
    to RGB.

    This affects JPEG transfer syntaxes.
    """
    return False


def supports_transfer_syntax(transfer_syntax: UID) -> bool:
    """Return ``True`` if the handler supports the `transfer_syntax`.

    Parameters
    ----------
    transfer_syntax : uid.UID
        The Transfer Syntax UID of the *Pixel Data* that is to be used with
        the handler.
    """
    return transfer_syntax in SUPPORTED_TRANSFER_SYNTAXES


def create_data_element(ds: "Dataset") -> "DataElement":
    """Return a ``gdcm.DataElement`` for the *Pixel Data*.

    Parameters
    ----------
    ds : dataset.Dataset
        The :class:`~pydicom.dataset.Dataset` containing the *Pixel
        Data*.

    Returns
    -------
    gdcm.DataElement
        The converted *Pixel Data* element.
    """
    tsyntax = ds.file_meta.TransferSyntaxUID
    data_element = gdcm.DataElement(gdcm.Tag(0x7fe0, 0x0010))
    if tsyntax.is_compressed:
        if getattr(ds, 'NumberOfFrames', 1) > 1:
            pixel_data_sequence = (
                pydicom.encaps.decode_data_sequence(ds.PixelData)
            )
        else:
            pixel_data_sequence = [
                pydicom.encaps.defragment_data(ds.PixelData)
            ]

        fragments = gdcm.SequenceOfFragments.New()
        for pixel_data in pixel_data_sequence:
            fragment = gdcm.Fragment()
            fragment.SetByteStringValue(pixel_data)
            fragments.AddFragment(fragment)
        data_element.SetValue(fragments.__ref__())
    else:
        data_element.SetByteStringValue(ds.PixelData)

    return data_element


def create_image(ds: "Dataset", data_element: "DataElement") -> "gdcm.Image":
    """Return a ``gdcm.Image``.

    Parameters
    ----------
    ds : dataset.Dataset
        The :class:`~pydicom.dataset.Dataset` containing the Image
        Pixel module.
    data_element : gdcm.DataElement
        The ``gdcm.DataElement`` *Pixel Data* element.

    Returns
    -------
    gdcm.Image
    """
    image = gdcm.Image()
    number_of_frames = getattr(ds, 'NumberOfFrames', 1)
    image.SetNumberOfDimensions(2 if number_of_frames == 1 else 3)
    image.SetDimensions((ds.Columns, ds.Rows, number_of_frames))
    image.SetDataElement(data_element)

    pi_type = gdcm.PhotometricInterpretation.GetPIType(
        ds.PhotometricInterpretation
    )
    image.SetPhotometricInterpretation(
        gdcm.PhotometricInterpretation(pi_type)
    )

    tsyntax = ds.file_meta.TransferSyntaxUID
    ts_type = gdcm.TransferSyntax.GetTSType(str.__str__(tsyntax))
    image.SetTransferSyntax(gdcm.TransferSyntax(ts_type))
    pixel_format = gdcm.PixelFormat(
        ds.SamplesPerPixel,
        ds.BitsAllocated,
        ds.BitsStored,
        ds.HighBit,
        ds.PixelRepresentation
    )
    image.SetPixelFormat(pixel_format)
    if 'PlanarConfiguration' in ds:
        image.SetPlanarConfiguration(ds.PlanarConfiguration)

    return image


def create_image_reader(ds: "Dataset") -> "gdcm.ImageReader":
    """Return a ``gdcm.ImageReader``.

    Parameters
    ----------
    ds : pydicom.dataset.Dataset
        The dataset to create the reader from.

    Returns
    -------
    gdcm.ImageReader
    """
    image_reader = gdcm.ImageReader()
    fname = getattr(ds, 'filename', None)
    if fname and isinstance(fname, str):
        pass
    else:
        # Copy the relevant elements and write to a temporary file to avoid
        #   having to deal with all the possible objects the dataset may
        #   originate with
        new = ds.group_dataset(0x0028)
        new["PixelData"] = ds["PixelData"]  # avoid ambiguous VR
        new.file_meta = ds.file_meta
        tfile = NamedTemporaryFile('wb')
        new.save_as(tfile.name)
        fname = tfile.name

    image_reader.SetFileName(fname)
    return image_reader


def get_pixeldata(ds: "Dataset") -> "numpy.ndarray":
    """Use the GDCM package to decode *Pixel Data*.

    Returns
    -------
    numpy.ndarray
        A correctly sized (but not shaped) array of the entire data volume

    Raises
    ------
    ImportError
        If the required packages are not available.
    TypeError
        If the image could not be read by GDCM or if the *Pixel Data* type is
        unsupported.
    AttributeError
        If the decoded amount of data does not match the expected amount.
    """
    if not HAVE_GDCM:
        raise ImportError("The GDCM handler requires both gdcm and numpy")

    if HAVE_GDCM_IN_MEMORY_SUPPORT:
        gdcm_data_element = create_data_element(ds)
        gdcm_image = create_image(ds, gdcm_data_element)
    else:
        gdcm_image_reader = create_image_reader(ds)
        if not gdcm_image_reader.Read():
            raise TypeError("GDCM could not read DICOM image")
        gdcm_image = gdcm_image_reader.GetImage()

    # GDCM returns char* as type str. Python 3 decodes this to
    # unicode strings by default.
    # The SWIG docs mention that they always decode byte streams
    # as utf-8 strings for Python 3, with the `surrogateescape`
    # error handler configured.
    # Therefore, we can encode them back to their original bytearray
    # representation on Python 3 by using the same parameters.

    pixel_bytearray = gdcm_image.GetBuffer().encode("utf-8", "surrogateescape")

    # Here we need to be careful because in some cases, GDCM reads a
    # buffer that is too large, so we need to make sure we only include
    # the first n_rows * n_columns * dtype_size bytes.
    expected_length_bytes = get_expected_length(ds)
    if ds.PhotometricInterpretation == 'YBR_FULL_422':
        # GDCM has already resampled the pixel data, see PS3.3 C.7.6.3.1.2
        expected_length_bytes = expected_length_bytes // 2 * 3

    if len(pixel_bytearray) > expected_length_bytes:
        # We make sure that all the bytes after are in fact zeros
        padding = pixel_bytearray[expected_length_bytes:]
        if numpy.any(numpy.frombuffer(padding, numpy.byte)):
            pixel_bytearray = pixel_bytearray[:expected_length_bytes]
        else:
            # We revert to the old behavior which should then result
            #   in a Numpy error later on.
            pass

    numpy_dtype = pixel_dtype(ds)
    arr = numpy.frombuffer(pixel_bytearray, dtype=numpy_dtype)

    expected_length_pixels = get_expected_length(ds, 'pixels')
    if arr.size != expected_length_pixels:
        raise AttributeError(
            f"Amount of pixel data {arr.size} does not match the "
            f"expected data {expected_length_pixels}"
        )

    tsyntax = ds.file_meta.TransferSyntaxUID
    if (
        config.APPLY_J2K_CORRECTIONS
        and tsyntax in [JPEG2000, JPEG2000Lossless]
    ):
        nr_frames = getattr(ds, 'NumberOfFrames', 1)
        codestream = next(generate_pixel_data(ds.PixelData, nr_frames))[0]

        params = get_j2k_parameters(codestream)
        j2k_precision = cast(
            int, params.setdefault("precision", ds.BitsStored)
        )
        j2k_sign = params.setdefault("is_signed", None)

        if not j2k_sign and ds.PixelRepresentation == 1:
            # Convert unsigned J2K data to 2's complement
            shift = cast(int, ds.BitsAllocated) - j2k_precision
            pixel_module = ds.group_dataset(0x0028)
            pixel_module.PixelRepresentation = 0
            dtype = pixel_dtype(pixel_module)
            arr = (arr.astype(dtype) << shift).astype(numpy_dtype) >> shift

    if should_change_PhotometricInterpretation_to_RGB(ds):
        ds.PhotometricInterpretation = "RGB"

    return cast("numpy.ndarray", arr.copy())



# Copyright 2008-2018 pydicom authors. See LICENSE file for details.
# -*- coding: utf-8 -*-
"""Unit tests for the pydicom.filereader module."""

import gzip
import io
from io import BytesIO
import os
import shutil
from pathlib import Path
from struct import unpack
import sys
import tempfile

import pytest

import pydicom.config
from pydicom import config
from pydicom.dataset import Dataset, FileDataset, FileMetaDataset
from pydicom.data import get_testdata_file
from pydicom.datadict import add_dict_entries
from pydicom.filereader import (
    dcmread, read_dataset, read_dicomdir, data_element_generator
)
from pydicom.dataelem import DataElement, DataElement_from_raw
from pydicom.errors import InvalidDicomError
from pydicom.filebase import DicomBytesIO
from pydicom.multival import MultiValue
from pydicom.sequence import Sequence
from pydicom.tag import Tag, TupleTag
from pydicom.uid import ImplicitVRLittleEndian
import pydicom.valuerep
from pydicom import values


from pydicom.pixel_data_handlers import gdcm_handler

have_gdcm_handler = gdcm_handler.is_available()

have_numpy = pydicom.config.have_numpy
if have_numpy:
    import numpy  # NOQA

try:
    import jpeg_ls
except ImportError:
    jpeg_ls = None

try:
    from PIL import Image as PILImg
except ImportError:
    # If that failed, try the alternate import syntax for PIL.
    try:
        import Image as PILImg
    except ImportError:
        # Neither worked, so it's likely not installed.
        PILImg = None

have_jpeg_ls = jpeg_ls is not None
have_pillow = PILImg is not None

empty_number_tags_name = get_testdata_file(
    "reportsi_with_empty_number_tags.dcm"
)
rtplan_name = get_testdata_file("rtplan.dcm")
rtdose_name = get_testdata_file("rtdose.dcm")
ct_name = get_testdata_file("CT_small.dcm")
mr_name = get_testdata_file("MR_small.dcm")
truncated_mr_name = get_testdata_file("MR_truncated.dcm")
jpeg2000_name = get_testdata_file("JPEG2000.dcm")
jpeg2000_embedded_sequence_delimeter_name = get_testdata_file(
    "JPEG2000-embedded-sequence-delimiter.dcm"
)
jpeg2000_lossless_name = get_testdata_file("MR_small_jp2klossless.dcm")
jpeg_ls_lossless_name = get_testdata_file("MR_small_jpeg_ls_lossless.dcm")
jpeg_lossy_name = get_testdata_file("JPEG-lossy.dcm")
jpeg_lossless_name = get_testdata_file("JPEG-LL.dcm")
deflate_name = get_testdata_file("image_dfl.dcm")
rtstruct_name = get_testdata_file("rtstruct.dcm")
priv_SQ_name = get_testdata_file("priv_SQ.dcm")
nested_priv_SQ_name = get_testdata_file("nested_priv_SQ.dcm")
meta_missing_tsyntax_name = get_testdata_file("meta_missing_tsyntax.dcm")
no_meta_group_length = get_testdata_file("no_meta_group_length.dcm")
gzip_name = get_testdata_file("zipMR.gz")
color_px_name = get_testdata_file("color-px.dcm")
color_pl_name = get_testdata_file("color-pl.dcm")
explicit_vr_le_no_meta = get_testdata_file("ExplVR_LitEndNoMeta.dcm")
explicit_vr_be_no_meta = get_testdata_file("ExplVR_BigEndNoMeta.dcm")
emri_name = get_testdata_file("emri_small.dcm")
emri_big_endian_name = get_testdata_file("emri_small_big_endian.dcm")
emri_jpeg_ls_lossless = get_testdata_file("emri_small_jpeg_ls_lossless.dcm")
emri_jpeg_2k_lossless = get_testdata_file("emri_small_jpeg_2k_lossless.dcm")
emri_jpeg_2k_lossless_too_short = get_testdata_file(
    "emri_small_jpeg_2k_lossless_too_short.dcm"
)
color_3d_jpeg_baseline = get_testdata_file("color3d_jpeg_baseline.dcm")
dir_name = os.path.dirname(sys.argv[0])
save_dir = os.getcwd()


class TestReader:
    def test_empty_numbers_tag(self):
        """Test that an empty tag with a number VR (FL, UL, SL, US,
        SS, FL, FD, OF) reads as ``None``."""
        empty_number_tags_ds = dcmread(empty_number_tags_name)
        assert empty_number_tags_ds.ExaminedBodyThickness is None
        assert empty_number_tags_ds.SimpleFrameList is None
        assert empty_number_tags_ds.ReferencePixelX0 is None
        assert empty_number_tags_ds.PhysicalUnitsXDirection is None
        assert empty_number_tags_ds.TagAngleSecondAxis is None
        assert empty_number_tags_ds.TagSpacingSecondDimension is None
        assert empty_number_tags_ds.VectorGridData is None

    def test_UTF8_filename(self):
        utf8_filename = os.path.join(tempfile.gettempdir(), "ДИКОМ.dcm")
        shutil.copyfile(rtdose_name, utf8_filename)
        ds = dcmread(utf8_filename)
        os.remove(utf8_filename)
        assert ds is not None

    def test_pathlib_path_filename(self):
        """Check that file can be read using pathlib.Path"""
        dcmread(Path(priv_SQ_name))

    def test_RTPlan(self):
        """Returns correct values for sample data elements in test
        RT Plan file.
        """
        plan = dcmread(rtplan_name)
        beam = plan.BeamSequence[0]
        # if not two controlpoints, then this would raise exception
        cp0, cp1 = beam.ControlPointSequence

        assert "unit001" == beam.TreatmentMachineName
        assert beam[0x300A, 0x00B2].value == beam.TreatmentMachineName

        got = cp1.ReferencedDoseReferenceSequence[
            0
        ].CumulativeDoseReferenceCoefficient
        DS = pydicom.valuerep.DS
        expected = DS("0.9990268")
        assert expected == got
        got = cp0.BeamLimitingDevicePositionSequence[0].LeafJawPositions
        if have_numpy and config.use_DS_numpy:
            expected = numpy.array([DS("-100"), DS("100.0")])
            assert numpy.allclose(got, expected)
        else:
            expected = [DS("-100"), DS("100.0")]
            assert got == expected

    def test_RTDose(self):
        """Returns correct values for sample data elements in test
        RT Dose file"""
        dose = dcmread(rtdose_name)
        assert Tag((0x3004, 0x000C)) == dose.FrameIncrementPointer
        assert dose[0x28, 9].value == dose.FrameIncrementPointer

        # try a value that is nested the deepest
        # (so deep I break it into two steps!)
        fract = dose.ReferencedRTPlanSequence[
            0
        ].ReferencedFractionGroupSequence[0]
        assert 1 == fract.ReferencedBeamSequence[0].ReferencedBeamNumber

    def test_CT(self):
        """Returns correct values for sample data elements in test CT file."""
        ct = dcmread(ct_name)
        assert "1.3.6.1.4.1.5962.2" == ct.file_meta.ImplementationClassUID
        value = ct.file_meta[0x2, 0x12].value
        assert value == ct.file_meta.ImplementationClassUID

        # (0020, 0032) Image Position (Patient)
        # [-158.13580300000001, -179.035797, -75.699996999999996]
        got = ct.ImagePositionPatient
        DS = pydicom.valuerep.DS
        if have_numpy and config.use_DS_numpy:
            expected = numpy.array([-158.135803, -179.035797, -75.699997])
            assert numpy.allclose(got, expected)
        else:
            expected = [DS("-158.135803"), DS("-179.035797"), DS("-75.699997")]
            assert got == expected

        assert 128 == ct.Rows
        assert 128 == ct.Columns
        assert 16 == ct.BitsStored
        assert 128 * 128 * 2 == len(ct.PixelData)

        # Also test private elements name can be resolved:
        got = ct[(0x0043, 0x104E)].name
        assert "[Duration of X-ray on]" == got

    @pytest.mark.skipif(not have_numpy, reason="Numpy not installed")
    def test_CT_PixelData(self):
        """Check that we can read pixel data.
        Tests that we get last one in array.
        """
        ct = dcmread(ct_name)
        assert 909 == ct.pixel_array[-1][-1]

    def test_no_force(self):
        """Raises exception if missing DICOM header and force==False."""
        with pytest.raises(InvalidDicomError):
            dcmread(rtstruct_name)

    def test_RTStruct(self):
        """Returns correct values for sample elements in test RTSTRUCT file."""
        # RTSTRUCT test file has complex nested sequences
        # -- see rtstruct.dump file
        # Also has no DICOM header ... so tests 'force' argument of dcmread

        rtss = dcmread(rtstruct_name, force=True)
        frame_of_ref = rtss.ReferencedFrameOfReferenceSequence[0]
        study = frame_of_ref.RTReferencedStudySequence[0]
        uid = study.RTReferencedSeriesSequence[0].SeriesInstanceUID
        assert "1.2.826.0.1.3680043.8.498.2010020400001.2.1.1" == uid

        got = rtss.ROIContourSequence[0].ContourSequence[2].ContourNumber
        assert 3 == got

        obs_seq0 = rtss.RTROIObservationsSequence[0]
        got = obs_seq0.ROIPhysicalPropertiesSequence[0].ROIPhysicalProperty
        assert "REL_ELEC_DENSITY" == got

    def test_dir(self):
        """Returns correct dir attributes for both Dataset and DICOM names
        (python >= 2.6).."""
        # Only python >= 2.6 calls __dir__ for dir() call
        rtss = dcmread(rtstruct_name, force=True)
        # sample some expected 'dir' values
        got_dir = dir(rtss)
        expect_in_dir = [
            "pixel_array",
            "add_new",
            "ROIContourSequence",
            "StructureSetDate",
        ]
        for name in expect_in_dir:
            assert name in got_dir

        # Now check for some items in dir() of a nested item
        roi0 = rtss.ROIContourSequence[0]
        got_dir = dir(roi0)
        expect_in_dir = [
            "pixel_array",
            "add_new",
            "ReferencedROINumber",
            "ROIDisplayColor",
        ]
        for name in expect_in_dir:
            assert name in got_dir

    def test_MR(self):
        """Returns correct values for sample data elements in test MR file."""
        mr = dcmread(mr_name)
        # (0010, 0010) Patient's Name           'CompressedSamples^MR1'
        mr.decode()
        assert "CompressedSamples^MR1" == mr.PatientName
        assert mr[0x10, 0x10].value == mr.PatientName

        DS = pydicom.valuerep.DS

        if have_numpy and config.use_DS_numpy:
            expected = numpy.array([0.3125, 0.3125])
            assert numpy.allclose(mr.PixelSpacing, expected)
        else:
            assert [DS("0.3125"), DS("0.3125")] == mr.PixelSpacing

    def test_deflate(self):
        """Returns correct values for sample data elements in test compressed
         (zlib deflate) file
         """
        # Everything after group 2 is compressed.
        # If we can read anything else, the decompression must have been ok.
        ds = dcmread(deflate_name)
        assert "WSD" == ds.ConversionType

    def test_sequence_with_implicit_vr(self):
        """Test that reading a UN sequence with unknown length and implicit VR
        in a dataset with explicit VR is read regardless of the value of
        the assume_implicit_vr_switch option."""
        replace_un_with_known_vr = config.replace_un_with_known_vr
        assume_implicit_vr_switch = config.assume_implicit_vr_switch

        config.replace_un_with_known_vr = True
        config.assume_implicit_vr_switch = True
        ds = dcmread(get_testdata_file("bad_sequence.dcm"))
        str(ds.CTDIPhantomTypeCodeSequence)

        config.assume_implicit_vr_switch = False
        ds = dcmread(get_testdata_file("bad_sequence.dcm"))
        str(ds.CTDIPhantomTypeCodeSequence)

        config.replace_un_with_known_vr = False
        ds = dcmread(get_testdata_file("bad_sequence.dcm"))
        str(ds.CTDIPhantomTypeCodeSequence)

        config.replace_un_with_known_vr = replace_un_with_known_vr
        config.assume_implicit_vr_switch = assume_implicit_vr_switch

    def test_no_pixels_read(self):
        """Returns all data elements before pixels using
        stop_before_pixels=False.
        """
        # Just check the tags, and a couple of values
        ctpartial = dcmread(ct_name, stop_before_pixels=True)
        ctpartial_tags = sorted(ctpartial.keys())
        ctfull = dcmread(ct_name)
        ctfull_tags = sorted(ctfull.keys())
        missing = [Tag(0x7FE0, 0x10), Tag(0xFFFC, 0xFFFC)]
        assert ctfull_tags == ctpartial_tags + missing

    @pytest.mark.skipif(not have_numpy, reason="Numpy not available")
    def test_no_float_pixels_read(self):
        """Returns all data elements before pixels using
        stop_before_pixels=True.
        """
        ds = Dataset()
        ds.InstanceNumber = 1
        ds.FloatPixelData = numpy.random.random((3, 3)).tobytes()

        fp = BytesIO()
        file_ds = FileDataset(fp, ds)
        file_ds.is_implicit_VR = True
        file_ds.is_little_endian = True
        file_ds.save_as(fp, write_like_original=True)

        test_ds = dcmread(fp, force=True, stop_before_pixels=True)
        ds_tags = sorted(ds.keys())
        test_ds_tags = sorted(test_ds.keys())
        assert ds_tags == test_ds_tags + [Tag(0x7FE0, 0x08)]

    def test_specific_tags(self):
        """Returns only tags specified by user."""
        ctspecific = dcmread(
            ct_name,
            specific_tags=[
                Tag(0x0010, 0x0010),
                "PatientID",
                "ImageType",
                "ViewName",
            ],
        )
        ctspecific_tags = sorted(ctspecific.keys())
        expected = [
            # SpecificCharacterSet is always added
            # ViewName does not exist in the data set
            Tag(0x0008, 0x0005),
            Tag(0x0008, 0x0008),
            Tag(0x0010, 0x0010),
            Tag(0x0010, 0x0020),
        ]
        assert expected == ctspecific_tags

    def test_specific_tags_with_other_unkonwn_length_tags(self):
        rtstruct_specific = dcmread(
            rtstruct_name,
            force=True,
            specific_tags=[
                "PatientName",
                "PatientID",
            ],
        )
        rtstruct_specific_tags = sorted(rtstruct_specific.keys())
        expected = [
            # SpecificCharacterSet is always added
            Tag(0x0008, 0x0005),
            Tag(0x0010, 0x0010),
            Tag(0x0010, 0x0020),
        ]
        assert expected == rtstruct_specific_tags

    def test_specific_tags_with_unknown_length_SQ(self):
        """Returns only tags specified by user."""
        unknown_len_sq_tag = Tag(0x3F03, 0x1001)
        tags = dcmread(priv_SQ_name, specific_tags=[unknown_len_sq_tag])
        tags = sorted(tags.keys())
        assert [unknown_len_sq_tag] == tags

        tags = dcmread(priv_SQ_name, specific_tags=["PatientName"])
        tags = sorted(tags.keys())
        assert [] == tags

    def test_specific_tags_with_unknown_length_tag(self):
        """Returns only tags specified by user."""
        unknown_len_tag = Tag(0x7FE0, 0x0010)  # Pixel Data
        tags = dcmread(emri_jpeg_2k_lossless, specific_tags=[unknown_len_tag])
        tags = sorted(tags.keys())
        # SpecificCharacterSet is always added
        assert [Tag(0x08, 0x05), unknown_len_tag] == tags

        tags = dcmread(
            emri_jpeg_2k_lossless, specific_tags=["SpecificCharacterSet"]
        )
        tags = sorted(tags.keys())
        assert [Tag(0x08, 0x05)] == tags

    def test_tag_with_unknown_length_tag_too_short(self, allow_invalid_values):
        """Tests handling of incomplete sequence value."""
        # the data set is the same as emri_jpeg_2k_lossless,
        # with the last 8 bytes removed to provoke the EOF error
        unknown_len_tag = Tag(0x7FE0, 0x0010)  # Pixel Data
        with pytest.warns(UserWarning, match="End of file reached*"):
            dcmread(
                emri_jpeg_2k_lossless_too_short,
                specific_tags=[unknown_len_tag],
            )

    def test_tag_with_unknown_length_tag_too_short_strict(
            self, enforce_valid_values):
        """Tests handling of incomplete sequence value in strict mode."""
        unknown_len_tag = Tag(0x7FE0, 0x0010)  # Pixel Data
        with pytest.raises(EOFError, match="End of file reached*"):
            dcmread(
                emri_jpeg_2k_lossless_too_short,
                specific_tags=[unknown_len_tag],
            )

    def test_private_SQ(self):
        """Can read private undefined length SQ without error."""
        # From issues 91, 97, 98. Bug introduced by fast reading, due to
        #    VR=None in raw data elements, then an undefined length private
        #    item VR is looked up, and there is no such tag,
        #    generating an exception

        # Simply read the file, in 0.9.5 this generated an exception
        dcmread(priv_SQ_name)

    def test_nested_private_SQ(self):
        """Can successfully read a private SQ which contains additional SQs."""
        # From issue 113. When a private SQ of undefined length is used, the
        #   sequence is read in and the length of the SQ is determined upon
        #   identification of the SQ termination sequence. When using nested
        #   Sequences, the first termination sequence encountered actually
        #   belongs to the nested Sequence not the parent, therefore the
        #   remainder of the file is not read in properly
        ds = dcmread(nested_priv_SQ_name)

        # Make sure that the entire dataset was read in
        pixel_data_tag = TupleTag((0x7FE0, 0x10))
        assert pixel_data_tag in ds

        # Check that the DataElement is indeed a Sequence
        tag = TupleTag((0x01, 0x01))
        seq0 = ds[tag]
        assert "SQ" == seq0.VR

        # Now verify the presence of the nested private SQ
        seq1 = seq0[0][tag]
        assert "SQ" == seq1.VR

        # Now make sure the values that are parsed are correct
        assert b"Double Nested SQ" == seq1[0][tag].value
        assert b"Nested SQ" == seq0[0][0x01, 0x02].value

    def test_un_sequence(self, dont_replace_un_with_known_vr):
        ds = dcmread(get_testdata_file("UN_sequence.dcm"))
        seq_element = ds[0x4453100c]
        assert seq_element.VR == "SQ"
        assert len(seq_element.value) == 1
        assert len(seq_element.value[0].ReferencedSeriesSequence) == 1

    def test_no_meta_group_length(self, no_datetime_conversion):
        """Read file with no group length in file meta."""
        # Issue 108 -- iView example file with no group length (0002,0002)
        # Originally crashed, now check no exception, but also check one item
        #     in file_meta, and second one in followinsg dataset
        ds = dcmread(no_meta_group_length)
        assert "20111130" == ds.InstanceCreationDate

    def test_no_transfer_syntax_in_meta(self):
        """Read file with file_meta, but has no TransferSyntaxUID in it."""
        # From issue 258: if file has file_meta but no TransferSyntaxUID in it,
        #   should assume default transfer syntax
        ds = dcmread(meta_missing_tsyntax_name)  # is default transfer syntax

        # Repeat one test from nested private sequence test to maker sure
        #    file was read correctly
        pixel_data_tag = TupleTag((0x7FE0, 0x10))
        assert pixel_data_tag in ds

    def test_explicit_VR_little_endian_no_meta(self, no_datetime_conversion):
        """Read file without file meta with Little Endian Explicit VR dataset.
        """
        # Example file from CMS XiO 5.0 and above
        # Still need to force read data since there is no 'DICM' marker present
        ds = dcmread(explicit_vr_le_no_meta, force=True)
        assert "20150529" == ds.InstanceCreationDate

    def test_explicit_VR_big_endian_no_meta(self, no_datetime_conversion):
        """Read file without file meta with Big Endian Explicit VR dataset."""
        # Example file from CMS XiO 5.0 and above
        # Still need to force read data since there is no 'DICM' marker present
        ds = dcmread(explicit_vr_be_no_meta, force=True)
        assert "20150529" == ds.InstanceCreationDate

    def test_planar_config(self):
        px_data_ds = dcmread(color_px_name)
        pl_data_ds = dcmread(color_pl_name)
        assert px_data_ds.PlanarConfiguration != pl_data_ds.PlanarConfiguration
        if have_numpy:
            px_data = px_data_ds.pixel_array
            pl_data = pl_data_ds.pixel_array
            assert numpy.all(px_data == pl_data)

    def test_correct_ambiguous_vr(self):
        """Test correcting ambiguous VR elements read from file"""
        ds = Dataset()
        ds.PixelRepresentation = 0
        ds.add(DataElement(0x00280108, "US", 10))
        ds.add(DataElement(0x00280109, "US", 500))

        fp = BytesIO()
        file_ds = FileDataset(fp, ds)
        file_ds.is_implicit_VR = True
        file_ds.is_little_endian = True
        file_ds.save_as(fp, write_like_original=True)

        ds = dcmread(fp, force=True)
        assert "US" == ds[0x00280108].VR
        assert 10 == ds.SmallestPixelValueInSeries

    def test_correct_ambiguous_explicit_vr(self):
        """Test correcting ambiguous VR elements read from file"""
        ds = Dataset()
        ds.PixelRepresentation = 0
        ds.add(DataElement(0x00280108, "US", 10))
        ds.add(DataElement(0x00280109, "US", 500))

        fp = BytesIO()
        file_ds = FileDataset(fp, ds)
        file_ds.is_implicit_VR = False
        file_ds.is_little_endian = True
        file_ds.save_as(fp, write_like_original=True)

        ds = dcmread(fp, force=True)
        assert "US" == ds[0x00280108].VR
        assert 10 == ds.SmallestPixelValueInSeries

    def test_correct_ambiguous_vr_compressed(self):
        """Test correcting compressed Pixel Data read from file"""
        # Create an implicit VR compressed dataset
        ds = dcmread(jpeg_lossless_name)
        fp = BytesIO()
        file_ds = FileDataset(fp, ds)
        file_ds.is_implicit_VR = True
        file_ds.is_little_endian = True
        file_ds.save_as(fp, write_like_original=True)

        ds = dcmread(fp, force=True)
        assert "OB" == ds[0x7FE00010].VR

    def test_read_encoded_pixel_data_without_embedded_sequence_delimiter(self):
        ds = dcmread(jpeg2000_name)
        assert "OB" == ds[0x7FE00010].VR
        assert 266 == len(ds[0x7FE00010].value)

    def test_read_encoded_pixel_data_with_embedded_sequence_delimiter(self):
        """Test ignoring embedded sequence delimiter in encoded pixel
        data fragment. Reproduces #1140.
        """
        ds = dcmread(jpeg2000_embedded_sequence_delimeter_name)
        assert "OB" == ds[0x7FE00010].VR
        assert 266 == len(ds[0x7FE00010].value)

    def test_long_specific_char_set(self, allow_invalid_values):
        """Test that specific character set is read even if it is longer
         than defer_size"""
        ds = Dataset()

        long_specific_char_set_value = ["ISO 2022IR 100"] * 9
        ds.add(DataElement(0x00080005, "CS", long_specific_char_set_value))

        msg = (
            r"Unknown encoding 'ISO 2022IR 100' - using default encoding "
            r"instead"
        )

        fp = BytesIO()
        file_ds = FileDataset(fp, ds)
        with pytest.warns(UserWarning, match=msg):
            file_ds.save_as(fp, write_like_original=True)

        with pytest.warns(UserWarning, match=msg):
            ds = dcmread(fp, defer_size=65, force=True)
            assert long_specific_char_set_value == ds[0x00080005].value

    def test_long_specific_char_set_strict(self, enforce_valid_values):
        ds = Dataset()

        long_specific_char_set_value = ["ISO 2022IR 100"] * 9
        ds.add(DataElement(0x00080005, "CS", long_specific_char_set_value))

        fp = BytesIO()
        file_ds = FileDataset(fp, ds)
        with pytest.raises(LookupError,
                           match="Unknown encoding 'ISO 2022IR 100'"):
            file_ds.save_as(fp, write_like_original=True)

    def test_no_preamble_file_meta_dataset(self):
        """Test correct read of group 2 elements with no preamble."""
        bytestream = (
            b"\x02\x00\x02\x00\x55\x49\x16\x00\x31\x2e\x32\x2e"
            b"\x38\x34\x30\x2e\x31\x30\x30\x30\x38\x2e\x35\x2e"
            b"\x31\x2e\x31\x2e\x39\x00\x02\x00\x10\x00\x55\x49"
            b"\x12\x00\x31\x2e\x32\x2e\x38\x34\x30\x2e\x31\x30"
            b"\x30\x30\x38\x2e\x31\x2e\x32\x00\x20\x20\x10\x00"
            b"\x02\x00\x00\x00\x01\x00\x20\x20\x20\x00\x06\x00"
            b"\x00\x00\x4e\x4f\x52\x4d\x41\x4c"
        )

        fp = BytesIO(bytestream)
        ds = dcmread(fp, force=True)
        assert "MediaStorageSOPClassUID" in ds.file_meta
        assert ImplicitVRLittleEndian == ds.file_meta.TransferSyntaxUID
        assert "NORMAL" == ds.Polarity
        assert 1 == ds.ImageBoxPosition

    def test_no_preamble_command_group_dataset(self):
        """Test correct read of group 0 and 2 elements with no preamble."""
        bytestream = (
            b"\x02\x00\x02\x00\x55\x49\x16\x00\x31\x2e\x32\x2e"
            b"\x38\x34\x30\x2e\x31\x30\x30\x30\x38\x2e\x35\x2e"
            b"\x31\x2e\x31\x2e\x39\x00\x02\x00\x10\x00\x55\x49"
            b"\x12\x00\x31\x2e\x32\x2e\x38\x34\x30\x2e\x31\x30"
            b"\x30\x30\x38\x2e\x31\x2e\x32\x00"
            b"\x20\x20\x10\x00\x02\x00\x00\x00\x01\x00\x20\x20"
            b"\x20\x00\x06\x00\x00\x00\x4e\x4f\x52\x4d\x41\x4c"
            b"\x00\x00\x10\x01\x02\x00\x00\x00\x03\x00"
        )

        fp = BytesIO(bytestream)
        ds = dcmread(fp, force=True)
        assert "MediaStorageSOPClassUID" in ds.file_meta
        assert ImplicitVRLittleEndian == ds.file_meta.TransferSyntaxUID
        assert "NORMAL" == ds.Polarity
        assert 1 == ds.ImageBoxPosition
        assert 3 == ds.MessageID

    def test_group_length_wrong(self):
        """Test file is read correctly even if FileMetaInformationGroupLength
        is incorrect.
        """
        bytestream = (
            b"\x02\x00\x00\x00\x55\x4C\x04\x00\x0A\x00\x00\x00"
            b"\x02\x00\x02\x00\x55\x49\x16\x00\x31\x2e\x32\x2e"
            b"\x38\x34\x30\x2e\x31\x30\x30\x30\x38\x2e\x35\x2e"
            b"\x31\x2e\x31\x2e\x39\x00\x02\x00\x10\x00\x55\x49"
            b"\x12\x00\x31\x2e\x32\x2e\x38\x34\x30\x2e\x31\x30"
            b"\x30\x30\x38\x2e\x31\x2e\x32\x00"
            b"\x20\x20\x10\x00\x02\x00\x00\x00\x01\x00\x20\x20"
            b"\x20\x00\x06\x00\x00\x00\x4e\x4f\x52\x4d\x41\x4c"
        )
        fp = BytesIO(bytestream)
        ds = dcmread(fp, force=True)
        value = ds.file_meta.FileMetaInformationGroupLength
        assert not len(bytestream) - 12 == value
        assert 10 == ds.file_meta.FileMetaInformationGroupLength
        assert "MediaStorageSOPClassUID" in ds.file_meta
        assert ImplicitVRLittleEndian == ds.file_meta.TransferSyntaxUID
        assert "NORMAL" == ds.Polarity
        assert 1 == ds.ImageBoxPosition

    def test_preamble_command_meta_no_dataset(self):
        """Test reading only preamble, command and meta elements"""
        preamble = b"\x00" * 128
        prefix = b"DICM"
        command = (
            b"\x00\x00\x00\x00\x04\x00\x00\x00\x38"
            b"\x00\x00\x00\x00\x00\x02\x00\x12\x00\x00"
            b"\x00\x31\x2e\x32\x2e\x38\x34\x30\x2e\x31"
            b"\x30\x30\x30\x38\x2e\x31\x2e\x31\x00\x00"
            b"\x00\x00\x01\x02\x00\x00\x00\x30\x00\x00"
            b"\x00\x10\x01\x02\x00\x00\x00\x07\x00\x00"
            b"\x00\x00\x08\x02\x00\x00\x00\x01\x01"
        )
        meta = (
            b"\x02\x00\x00\x00\x55\x4C\x04\x00\x0A\x00\x00\x00"
            b"\x02\x00\x02\x00\x55\x49\x16\x00\x31\x2e\x32\x2e"
            b"\x38\x34\x30\x2e\x31\x30\x30\x30\x38\x2e\x35\x2e"
            b"\x31\x2e\x31\x2e\x39\x00\x02\x00\x10\x00\x55\x49"
            b"\x12\x00\x31\x2e\x32\x2e\x38\x34\x30\x2e\x31\x30"
            b"\x30\x30\x38\x2e\x31\x2e\x32\x00"
        )

        bytestream = preamble + prefix + meta + command
        fp = BytesIO(bytestream)
        ds = dcmread(fp, force=True)
        assert "TransferSyntaxUID" in ds.file_meta
        assert "MessageID" in ds

    def test_preamble_meta_no_dataset(self):
        """Test reading only preamble and meta elements"""
        preamble = b"\x00" * 128
        prefix = b"DICM"
        meta = (
            b"\x02\x00\x00\x00\x55\x4C\x04\x00\x0A\x00\x00\x00"
            b"\x02\x00\x02\x00\x55\x49\x16\x00\x31\x2e\x32\x2e"
            b"\x38\x34\x30\x2e\x31\x30\x30\x30\x38\x2e\x35\x2e"
            b"\x31\x2e\x31\x2e\x39\x00\x02\x00\x10\x00\x55\x49"
            b"\x12\x00\x31\x2e\x32\x2e\x38\x34\x30\x2e\x31\x30"
            b"\x30\x30\x38\x2e\x31\x2e\x32\x00"
        )

        bytestream = preamble + prefix + meta
        fp = BytesIO(bytestream)
        ds = dcmread(fp, force=True)
        assert b"\x00" * 128 == ds.preamble
        assert "TransferSyntaxUID" in ds.file_meta
        assert Dataset() == ds[:]

    def test_preamble_commandset_no_dataset(self):
        """Test reading only preamble and command set"""
        preamble = b"\x00" * 128
        prefix = b"DICM"
        command = (
            b"\x00\x00\x00\x00\x04\x00\x00\x00\x38"
            b"\x00\x00\x00\x00\x00\x02\x00\x12\x00\x00"
            b"\x00\x31\x2e\x32\x2e\x38\x34\x30\x2e\x31"
            b"\x30\x30\x30\x38\x2e\x31\x2e\x31\x00\x00"
            b"\x00\x00\x01\x02\x00\x00\x00\x30\x00\x00"
            b"\x00\x10\x01\x02\x00\x00\x00\x07\x00\x00"
            b"\x00\x00\x08\x02\x00\x00\x00\x01\x01"
        )
        bytestream = preamble + prefix + command

        fp = BytesIO(bytestream)
        ds = dcmread(fp, force=True)
        assert "MessageID" in ds
        assert Dataset() == ds.file_meta

    def test_meta_no_dataset(self):
        """Test reading only meta elements"""
        bytestream = (
            b"\x02\x00\x00\x00\x55\x4C\x04\x00\x0A\x00\x00\x00"
            b"\x02\x00\x02\x00\x55\x49\x16\x00\x31\x2e\x32\x2e"
            b"\x38\x34\x30\x2e\x31\x30\x30\x30\x38\x2e\x35\x2e"
            b"\x31\x2e\x31\x2e\x39\x00\x02\x00\x10\x00\x55\x49"
            b"\x12\x00\x31\x2e\x32\x2e\x38\x34\x30\x2e\x31\x30"
            b"\x30\x30\x38\x2e\x31\x2e\x32\x00"
        )
        fp = BytesIO(bytestream)
        ds = dcmread(fp, force=True)
        assert "TransferSyntaxUID" in ds.file_meta
        assert Dataset() == ds[:]

    def test_commandset_no_dataset(self):
        """Test reading only command set elements"""
        bytestream = (
            b"\x00\x00\x00\x00\x04\x00\x00\x00\x38"
            b"\x00\x00\x00\x00\x00\x02\x00\x12\x00\x00"
            b"\x00\x31\x2e\x32\x2e\x38\x34\x30\x2e\x31"
            b"\x30\x30\x30\x38\x2e\x31\x2e\x31\x00\x00"
            b"\x00\x00\x01\x02\x00\x00\x00\x30\x00\x00"
            b"\x00\x10\x01\x02\x00\x00\x00\x07\x00\x00"
            b"\x00\x00\x08\x02\x00\x00\x00\x01\x01"
        )
        fp = BytesIO(bytestream)
        ds = dcmread(fp, force=True)
        assert "MessageID" in ds
        assert ds.preamble is None
        assert Dataset() == ds.file_meta

    def test_file_meta_dataset_implicit_vr(self, allow_invalid_values):
        """Test reading a file meta dataset that is implicit VR"""

        bytestream = (
            b"\x02\x00\x10\x00\x12\x00\x00\x00"
            b"\x31\x2e\x32\x2e\x38\x34\x30\x2e"
            b"\x31\x30\x30\x30\x38\x2e\x31\x2e"
            b"\x32\x00"
        )
        fp = BytesIO(bytestream)
        with pytest.warns(UserWarning):
            ds = dcmread(fp, force=True)
        assert "TransferSyntaxUID" in ds.file_meta

    def test_file_meta_dataset_implicit_vr_strict(self, enforce_valid_values):
        """Test reading a file meta dataset that is implicit VR"""

        bytestream = (
            b"\x02\x00\x10\x00\x12\x00\x00\x00"
            b"\x31\x2e\x32\x2e\x38\x34\x30\x2e"
            b"\x31\x30\x30\x30\x38\x2e\x31\x2e"
            b"\x32\x00"
        )
        fp = BytesIO(bytestream)
        with pytest.raises(InvalidDicomError,
                           match="Expected explicit VR, "
                                 "but found implicit VR"):
            dcmread(fp, force=True)

    def test_no_dataset(self):
        """Test reading no elements or preamble produces empty Dataset"""
        bytestream = b""
        fp = BytesIO(bytestream)
        ds = dcmread(fp, force=True)
        assert ds.preamble is None
        assert Dataset() == ds.file_meta
        assert Dataset() == ds[:]

    def test_empty_file(self):
        """Test reading no elements from file produces empty Dataset"""
        with tempfile.NamedTemporaryFile() as f:
            ds = dcmread(f, force=True)
            assert ds.preamble is None
            assert Dataset() == ds.file_meta
            assert Dataset() == ds[:]

    def test_bad_filename(self):
        """Test reading from non-existing file raises."""
        with pytest.raises(FileNotFoundError):
            dcmread("InvalidFilePath")
        with pytest.raises(TypeError, match="dcmread: Expected a file path or "
                                            "a file-like, but got None"):
            dcmread(None)
        with pytest.raises(TypeError, match="dcmread: Expected a file path or "
                                            "a file-like, but got int"):
            dcmread(42)

    def test_empty_specific_character_set(self):
        """Test that an empty Specific Character Set is handled correctly.
        Regression test for #1038"""
        ds = dcmread(get_testdata_file("empty_charset_LEI.dcm"))
        assert ds.read_encoding == ["iso8859"]

    def test_dcmread_does_not_raise(self):
        """Test that reading from DicomBytesIO does not raise on EOF.
        Regression test for #358."""
        ds = dcmread(mr_name)
        fp = DicomBytesIO()
        ds.save_as(fp, write_like_original=True)
        fp.seek(0)
        de_gen = data_element_generator(fp, False, True)
        try:
            while True:
                next(de_gen)
        except StopIteration:
            pass
        except EOFError:
            self.fail("Unexpected EOFError raised")

    def test_lut_descriptor(self):
        """Regression test for #942: incorrect first value"""
        prefixes = [
            b"\x28\x00\x01\x11",
            b"\x28\x00\x02\x11",
            b"\x28\x00\x03\x11",
            b"\x28\x00\x02\x30",
        ]
        suffix = b"\x53\x53\x06\x00\x00\xf5\x00\xf8\x10\x00"

        for raw_tag in prefixes:
            tag = unpack("<2H", raw_tag)
            bs = DicomBytesIO(raw_tag + suffix)
            bs.is_little_endian = True
            bs.is_implicit_VR = False

            ds = dcmread(bs, force=True)
            elem = ds[tag]
            assert elem.VR == "SS"
            assert elem.value == [62720, -2048, 16]

    def test_lut_descriptor_empty(self):
        """Regression test for #1049: LUT empty raises."""
        bs = DicomBytesIO(b"\x28\x00\x01\x11\x53\x53\x00\x00")
        bs.is_little_endian = True
        bs.is_implicit_VR = False
        ds = dcmread(bs, force=True)
        elem = ds[0x00281101]
        assert elem.value is None
        assert elem.VR == "SS"

    def test_lut_descriptor_singleton(self):
        """Test LUT Descriptor with VM = 1"""
        bs = DicomBytesIO(b"\x28\x00\x01\x11\x53\x53\x02\x00\x00\xf5")
        bs.is_little_endian = True
        bs.is_implicit_VR = False
        ds = dcmread(bs, force=True)
        elem = ds[0x00281101]
        # No conversion to US if not a triplet
        assert elem.value == -2816
        assert elem.VR == "SS"

    def test_reading_of(self):
        """Test reading a dataset with OF element."""
        bs = DicomBytesIO(
            b"\x28\x00\x01\x11\x53\x53\x06\x00\x00\xf5\x00\xf8\x10\x00"
            b"\xe0\x7f\x08\x00\x4F\x46\x00\x00\x04\x00\x00\x00\x00\x01\x02\x03"
        )
        bs.is_little_endian = True
        bs.is_implicit_VR = False

        ds = dcmread(bs, force=True)
        elem = ds["FloatPixelData"]
        assert "OF" == elem.VR
        assert b"\x00\x01\x02\x03" == elem.value

    def test_empty_pn(self):
        """Test correct type for an empty PN element."""
        # Test for 1338
        ds = Dataset()
        ds.is_little_endian = True
        ds.is_implicit_VR = True
        ds.PatientName = ''
        assert isinstance(ds.PatientName, pydicom.valuerep.PersonName)

        bs = DicomBytesIO()
        ds.save_as(bs)

        out = dcmread(bs, force=True)
        assert isinstance(ds[0x00100010].value, pydicom.valuerep.PersonName)


class TestIncorrectVR:
    def setup(self):
        self.ds_explicit = BytesIO(
            b"\x08\x00\x05\x00CS\x0a\x00ISO_IR 100"  # SpecificCharacterSet
            b"\x08\x00\x20\x00DA\x08\x0020000101"  # StudyDate
        )
        self.ds_implicit = BytesIO(
            b"\x08\x00\x05\x00\x0a\x00\x00\x00ISO_IR 100"
            b"\x08\x00\x20\x00\x08\x00\x00\x0020000101"
        )

    def test_implicit_vr_expected_explicit_used(self, allow_invalid_values,
                                                no_datetime_conversion):
        msg = (
            "Expected implicit VR, but found explicit VR - "
            "using explicit VR for reading"
        )

        with pytest.warns(UserWarning, match=msg):
            ds = read_dataset(
                self.ds_explicit, is_implicit_VR=True, is_little_endian=True
            )
        assert "ISO_IR 100" == ds.SpecificCharacterSet
        assert "20000101" == ds.StudyDate

    def test_implicit_vr_expected_explicit_used_strict(
            self, enforce_valid_values):
        msg = "Expected implicit VR, but found explicit VR"
        with pytest.raises(InvalidDicomError, match=msg):
            read_dataset(
                self.ds_explicit, is_implicit_VR=True, is_little_endian=True
            )

    def test_explicit_vr_expected_implicit_used(self, allow_invalid_values,
                                                no_datetime_conversion):
        msg = (
            "Expected explicit VR, but found implicit VR - "
            "using implicit VR for reading"
        )

        with pytest.warns(UserWarning, match=msg):
            ds = read_dataset(
                self.ds_implicit, is_implicit_VR=False, is_little_endian=True
            )
        assert "ISO_IR 100" == ds.SpecificCharacterSet
        assert "20000101" == ds.StudyDate

    def test_explicit_vr_expected_implicit_used_strict(
            self, enforce_valid_values):
        msg = "Expected explicit VR, but found implicit VR"
        with pytest.raises(InvalidDicomError, match=msg):
            read_dataset(
                self.ds_implicit, is_implicit_VR=False, is_little_endian=True
            )

    def test_seq_item_looks_like_explicit_VR(self):
        # For issue 999.

        # Set up an implicit VR dataset with a "normal" group 8 tag,
        # followed by a sequence with an item (dataset) having
        # a data element length that looks like a potential valid VR
        ds = Dataset()
        ds.file_meta = FileMetaDataset()
        ds.file_meta.MediaStorageSOPClassUID = "1.1.1"
        ds.file_meta.MediaStorageSOPInstanceUID = "2.2.2"
        ds.is_implicit_VR = True
        ds.is_little_endian = True
        ds.SOPClassUID = "9.9.9"  # First item group 8 in top-level dataset
        seq = Sequence()
        seq_ds = Dataset()
        seq_ds.BadPixelImage = b"\3" * 0x5244  # length looks like "DR"
        seq.append(seq_ds)
        ds.ReferencedImageSequence = seq

        dbio = DicomBytesIO()
        ds.save_as(dbio, write_like_original=False)

        # Now read the constructed dataset back in
        # In original issue, shows warning that has detected what appears
        # to be Explicit VR, then throws NotImplemented for the unknown VR
        dbio.seek(0)
        ds = dcmread(dbio)
        ds.remove_private_tags()  # forces it to actually parse SQ


class TestUnknownVR:
    @pytest.fixture(autouse=True)
    def restore_config_values(self):
        orig_impl_VR_switch = config.assume_implicit_vr_switch
        config.assume_implicit_vr_switch = False
        yield
        config.assume_implicit_vr_switch = orig_impl_VR_switch

    @pytest.mark.parametrize(
        "vr_bytes, str_output",
        [
            # Test limits of char values
            (b"\x00\x41", "0x00 0x41"),  # 000/A
            (b"\x40\x41", "0x40 0x41"),  # 064/A
            (b"\x5B\x41", "0x5b 0x41"),  # 091/A
            (b"\x60\x41", "0x60 0x41"),  # 096/A
            (b"\x7B\x41", "0x7b 0x41"),  # 123/A
            (b"\xFF\x41", "0xff 0x41"),  # 255/A
            # Test good/bad
            (b"\x41\x00", "0x41 0x00"),  # A/-
            (b"\x5A\x00", "0x5a 0x00"),  # Z/-
            # Test not quite good/bad
            (b"\x61\x00", "0x61 0x00"),  # a/-
            (b"\x7A\x00", "0x7a 0x00"),  # z/-
            # Test bad/good
            (b"\x00\x41", "0x00 0x41"),  # -/A
            (b"\x00\x5A", "0x00 0x5a"),  # -/Z
            # Test bad/not quite good
            (b"\x00\x61", "0x00 0x61"),  # -/a
            (b"\x00\x7A", "0x00 0x7a"),  # -/z
            # Test good/good
            (b"\x41\x41", "AA"),  # A/A
            (b"\x41\x5A", "AZ"),  # A/Z
            (b"\x5A\x41", "ZA"),  # Z/A
            (b"\x5A\x5A", "ZZ"),  # Z/Z
            # Test not quite good
            (b"\x41\x61", "Aa"),  # A/a
            (b"\x41\x7A", "Az"),  # A/z
            (b"\x61\x41", "aA"),  # a/A
            (b"\x61\x5A", "aZ"),  # a/Z
            (b"\x61\x61", "aa"),  # a/a
            (b"\x61\x7A", "az"),  # a/z
            (b"\x5A\x61", "Za"),  # Z/a
            (b"\x5A\x7A", "Zz"),  # Z/z
            (b"\x7A\x41", "zA"),  # z/A
            (b"\x7A\x5A", "zZ"),  # z/Z
            (b"\x7A\x61", "za"),  # z/a
            (b"\x7A\x7A", "zz"),  # z/z
        ],
    )
    def test_fail_decode_msg(self, vr_bytes, str_output):
        """Regression test for #791."""
        # start the dataset with a valid tag (SpecificCharacterSet),
        # as the first tag is used to check the VR
        ds = read_dataset(
            BytesIO(
                b"\x08\x00\x05\x00CS\x0a\x00ISO_IR 100"
                b"\x08\x00\x06\x00" + vr_bytes + b"\x00\x00\x00\x08\x00\x49"
            ),
            False,
            True,
        )
        msg = r"Unknown Value Representation '{}' in tag \(0008, 0006\)"
        msg = msg.format(str_output)
        with pytest.raises(NotImplementedError, match=msg):
            print(ds)


class TestReadDataElement:
    def setup(self):
        ds = Dataset()
        ds.DoubleFloatPixelData = (
            b"\x00\x01\x02\x03\x04\x05\x06\x07"
            b"\x01\x01\x02\x03\x04\x05\x06\x07"
        )  # OD
        ds.SelectorOLValue = (
            b"\x00\x01\x02\x03\x04\x05\x06\x07" b"\x01\x01\x02\x03"
        )  # VR of OL
        ds.PotentialReasonsForProcedure = [
            "A",
            "B",
            "C",
        ]  # VR of UC, odd length
        ds.StrainDescription = "Test"  # Even length
        ds.URNCodeValue = "http://test.com"  # VR of UR
        ds.RetrieveURL = "ftp://test.com  "  # Test trailing spaces ignored
        ds.DestinationAE = "    TEST  12    "  # 16 characters max for AE
        # 8-byte values
        ds.ExtendedOffsetTable = (  # VR of OV
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x01\x02\x03\x04\x05\x06\x07\x08"
        )

        # No public elements with VR of SV or UV yet...
        add_dict_entries(
            {
                0xFFFE0001: (
                    "SV",
                    "1",
                    "SV Element Minimum",
                    "",
                    "SVElementMinimum",
                ),
                0xFFFE0002: (
                    "SV",
                    "1",
                    "SV Element Maximum",
                    "",
                    "SVElementMaximum",
                ),
                0xFFFE0003: (
                    "UV",
                    "1",
                    "UV Element Minimum",
                    "",
                    "UVElementMinimum",
                ),
                0xFFFE0004: (
                    "UV",
                    "1",
                    "UV Element Maximum",
                    "",
                    "UVElementMaximum",
                ),
            }
        )
        ds.SVElementMinimum = -(2 ** 63)
        ds.SVElementMaximum = 2 ** 63 - 1
        ds.UVElementMinimum = 0
        ds.UVElementMaximum = 2 ** 64 - 1

        self.fp = BytesIO()  # Implicit little
        file_ds = FileDataset(self.fp, ds)
        file_ds.is_implicit_VR = True
        file_ds.is_little_endian = True
        file_ds.save_as(self.fp, write_like_original=True)

        self.fp_ex = BytesIO()  # Explicit little
        file_ds = FileDataset(self.fp_ex, ds)
        file_ds.is_implicit_VR = False
        file_ds.is_little_endian = True
        file_ds.save_as(self.fp_ex, write_like_original=True)

    def test_read_OD_implicit_little(self):
        """Check creation of OD DataElement from byte data works correctly."""
        ds = dcmread(self.fp, force=True)
        ref_elem = ds.get(0x7FE00009)
        elem = DataElement(
            0x7FE00009,
            "OD",
            b"\x00\x01\x02\x03\x04\x05\x06\x07"
            b"\x01\x01\x02\x03\x04\x05\x06\x07",
        )
        assert ref_elem == elem

    def test_read_OD_explicit_little(self):
        """Check creation of OD DataElement from byte data works correctly."""
        ds = dcmread(self.fp_ex, force=True)
        ref_elem = ds.get(0x7FE00009)
        elem = DataElement(
            0x7FE00009,
            "OD",
            b"\x00\x01\x02\x03\x04\x05\x06\x07"
            b"\x01\x01\x02\x03\x04\x05\x06\x07",
        )
        assert ref_elem == elem

    def test_read_OL_implicit_little(self):
        """Check creation of OL DataElement from byte data works correctly."""
        ds = dcmread(self.fp, force=True)
        ref_elem = ds.get(0x00720075)
        elem = DataElement(
            0x00720075,
            "OL",
            b"\x00\x01\x02\x03\x04\x05\x06\x07" b"\x01\x01\x02\x03",
        )
        assert ref_elem == elem

    def test_read_OL_explicit_little(self):
        """Check creation of OL DataElement from byte data works correctly."""
        ds = dcmread(self.fp_ex, force=True)
        ref_elem = ds.get(0x00720075)
        elem = DataElement(
            0x00720075,
            "OL",
            b"\x00\x01\x02\x03\x04\x05\x06\x07" b"\x01\x01\x02\x03",
        )
        assert ref_elem == elem

    def test_read_UC_implicit_little(self):
        """Check creation of DataElement from byte data works correctly."""
        ds = dcmread(self.fp, force=True)
        ref_elem = ds.get(0x00189908)
        elem = DataElement(0x00189908, "UC", ["A", "B", "C"])
        assert ref_elem == elem

        ds = dcmread(self.fp, force=True)
        ref_elem = ds.get(0x00100212)
        elem = DataElement(0x00100212, "UC", "Test")
        assert ref_elem == elem

    def test_read_UC_explicit_little(self):
        """Check creation of DataElement from byte data works correctly."""
        ds = dcmread(self.fp_ex, force=True)
        ref_elem = ds.get(0x00189908)
        elem = DataElement(0x00189908, "UC", ["A", "B", "C"])
        assert ref_elem == elem

        ds = dcmread(self.fp_ex, force=True)
        ref_elem = ds.get(0x00100212)
        elem = DataElement(0x00100212, "UC", "Test")
        assert ref_elem == elem

    def test_read_UR_implicit_little(self):
        """Check creation of DataElement from byte data works correctly."""
        ds = dcmread(self.fp, force=True)
        ref_elem = ds.get(0x00080120)  # URNCodeValue
        elem = DataElement(0x00080120, "UR", "http://test.com")
        assert ref_elem == elem

        # Test trailing spaces ignored
        ref_elem = ds.get(0x00081190)  # RetrieveURL
        elem = DataElement(0x00081190, "UR", "ftp://test.com")
        assert ref_elem == elem

    def test_read_UR_explicit_little(self):
        """Check creation of DataElement from byte data works correctly."""
        ds = dcmread(self.fp_ex, force=True)
        ref_elem = ds.get(0x00080120)  # URNCodeValue
        elem = DataElement(0x00080120, "UR", "http://test.com")
        assert ref_elem == elem

        # Test trailing spaces ignored
        ref_elem = ds.get(0x00081190)  # RetrieveURL
        elem = DataElement(0x00081190, "UR", "ftp://test.com")
        assert ref_elem == elem

    def test_read_AE(self):
        """Check creation of AE DataElement from byte data works correctly."""
        ds = dcmread(self.fp, force=True)
        assert "TEST  12" == ds.DestinationAE

        # Test multivalue read correctly
        ds.DestinationAE = ["TEST  12  ", "  TEST2", "   TEST 3  "]

        fp = BytesIO()
        ds.save_as(fp, write_like_original=True)
        fp.seek(0)
        ds = dcmread(fp, force=True)
        assert ["TEST  12", "TEST2", "TEST 3"] == ds.DestinationAE

    def test_read_OV_implicit_little(self):
        """Check reading element with VR of OV encoded as implicit"""
        ds = dcmread(self.fp, force=True)
        val = (
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x01\x02\x03\x04\x05\x06\x07\x08"
        )
        elem = ds["ExtendedOffsetTable"]
        assert "OV" == elem.VR
        assert 0x7FE00001 == elem.tag
        assert val == elem.value

        new = DataElement(0x7FE00001, "OV", val)
        assert elem == new

    def test_read_OV_explicit_little(self):
        """Check reading element with VR of OV encoded as explicit"""
        ds = dcmread(self.fp_ex, force=True)
        val = (
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x01\x02\x03\x04\x05\x06\x07\x08"
        )
        elem = ds["ExtendedOffsetTable"]
        assert "OV" == elem.VR
        assert 0x7FE00001 == elem.tag
        assert val == elem.value

        new = DataElement(0x7FE00001, "OV", val)
        assert elem == new

    def test_read_SV_implicit_little(self):
        """Check reading element with VR of SV encoded as implicit"""
        ds = dcmread(self.fp, force=True)
        elem = ds["SVElementMinimum"]
        assert "SV" == elem.VR
        assert 0xFFFE0001 == elem.tag
        assert -(2 ** 63) == elem.value

        new = DataElement(0xFFFE0001, "SV", -(2 ** 63))
        assert elem == new

        elem = ds["SVElementMaximum"]
        assert "SV" == elem.VR
        assert 0xFFFE0002 == elem.tag
        assert 2 ** 63 - 1 == elem.value

        new = DataElement(0xFFFE0002, "SV", 2 ** 63 - 1)
        assert elem == new

    @pytest.mark.skip("No public elements with VR of SV")
    def test_read_SV_explicit_little(self):
        """Check reading element with VR of SV encoded as explicit"""
        ds = dcmread(self.fp_ex, force=True)
        elem = ds["SVElementMinimum"]
        assert "SV" == elem.VR
        assert 0xFFFE0001 == elem.tag
        assert -(2 ** 63) == elem.value

        new = DataElement(0xFFFE0001, "SV", -(2 ** 63))
        assert elem == new

        elem = ds["SVElementMaximum"]
        assert "SV" == elem.VR
        assert 0xFFFE0002 == elem.tag
        assert 2 ** 63 - 1 == elem.value

        new = DataElement(0xFFFE0002, "SV", 2 ** 63 - 1)
        assert elem == new

    def test_read_UV_implicit_little(self):
        """Check reading element with VR of UV encoded as implicit"""
        ds = dcmread(self.fp, force=True)
        elem = ds["UVElementMinimum"]
        assert "UV" == elem.VR
        assert 0xFFFE0003 == elem.tag
        assert 0 == elem.value

        new = DataElement(0xFFFE0003, "UV", 0)
        assert elem == new

        elem = ds["UVElementMaximum"]
        assert "UV" == elem.VR
        assert 0xFFFE0004 == elem.tag
        assert 2 ** 64 - 1 == elem.value

        new = DataElement(0xFFFE0004, "UV", 2 ** 64 - 1)
        assert elem == new

    def test_read_UV_explicit_little(self):
        """Check reading element with VR of UV encoded as explicit"""
        ds = dcmread(self.fp_ex, force=True)
        elem = ds["UVElementMinimum"]
        assert "UV" == elem.VR
        assert 0xFFFE0003 == elem.tag
        assert 0 == elem.value

        new = DataElement(0xFFFE0003, "UV", 0)
        assert elem == new

        elem = ds["UVElementMaximum"]
        assert "UV" == elem.VR
        assert 0xFFFE0004 == elem.tag
        assert 2 ** 64 - 1 == elem.value

        new = DataElement(0xFFFE0004, "UV", 2 ** 64 - 1)
        assert elem == new


class TestDSISnumpy:
    @pytest.fixture(autouse=True)
    def restore_config_values(self):
        orig_IS_numpy = config.use_IS_numpy
        orig_DS_numpy = config.use_DS_numpy
        orig_DS_decimal = config.use_DS_decimal
        yield
        config.use_IS_numpy = orig_IS_numpy
        config.DS_decimal(orig_DS_decimal)
        config.DS_numpy(orig_DS_numpy)

    @pytest.mark.skipif(have_numpy, reason="Testing import error")
    def test_IS_numpy_import_error(self):
        config.use_IS_numpy = True
        rtss = dcmread(rtstruct_name, force=True)
        # no numpy, then trying to use numpy raises error
        with pytest.raises(ImportError):
            rtss.ROIContourSequence[0].ROIDisplayColor  # VR is IS

    @pytest.mark.skipif(not have_numpy, reason="Testing with numpy only")
    def test_IS_numpy_class(self):
        config.use_IS_numpy = True
        rtss = dcmread(rtstruct_name, force=True)
        col = rtss.ROIContourSequence[0].ROIDisplayColor  # VR is IS
        assert isinstance(col, numpy.ndarray)
        assert "int64" == col.dtype

        # Check a conversion with only a single value
        roi_num = rtss.ROIContourSequence[0].ReferencedROINumber
        assert isinstance(roi_num, numpy.int64)

    def test_IS_not_numpy(self):
        """Test class of the object matches the config,
        when the config is changed"""
        config.use_IS_numpy = False
        rtss = dcmread(rtstruct_name, force=True)
        col = rtss.ROIContourSequence[0].ROIDisplayColor  # VR is IS
        assert isinstance(col, MultiValue)

    @pytest.mark.skipif(have_numpy, reason="Testing import error")
    def test_DS_numpy_import_error(self):
        config.use_DS_numpy = True
        rtss = dcmread(rtstruct_name, force=True)
        # no numpy, then trying to use numpy raises error
        with pytest.raises(ImportError):
            rtss.ROIContourSequence[0].ContourSequence[0].ContourData

    @pytest.mark.skipif(not have_numpy, reason="Testing with numpy only")
    def test_DS_numpy_class(self):
        config.use_DS_numpy = True
        rtss = dcmread(rtstruct_name, force=True)
        # ContourData has VR of DS
        cd = rtss.ROIContourSequence[0].ContourSequence[0].ContourData
        assert isinstance(cd, numpy.ndarray)
        assert "float64" == cd.dtype

        # Check conversion with only a single value
        roi_vol = rtss.StructureSetROISequence[0].ROIVolume
        assert isinstance(roi_vol, numpy.float64)

    def test_DS_not_numpy(self):
        """Test class of the object matches the config."""
        config.use_DS_numpy = False
        rtss = dcmread(rtstruct_name, force=True)
        # ContourData has VR of DS
        cd = rtss.ROIContourSequence[0].ContourSequence[0].ContourData
        assert isinstance(cd, MultiValue)

    @pytest.mark.skipif(not have_numpy, reason="numpy not installed")
    def test_DS_conflict_config(self):
        config.DS_numpy(False)
        config.DS_decimal(True)
        with pytest.raises(ValueError):
            config.DS_numpy(True)

    @pytest.mark.skipif(not have_numpy, reason="numpy not installed")
    def test_DS_conflict_config2(self):
        config.DS_numpy(True)
        with pytest.raises(ValueError):
            config.DS_decimal(True)

    @pytest.mark.skipif(not have_numpy, reason="numpy not installed")
    def test_DS_bad_chars(self):
        config.DS_numpy(True)
        with pytest.raises(ValueError):
            values.convert_DS_string(b"123.1b", True)

    @pytest.mark.skipif(not have_numpy, reason="numpy not installed")
    def test_IS_bad_chars(self):
        config.use_IS_numpy = True
        with pytest.raises(ValueError):
            values.convert_IS_string(b"123b", True)

    @pytest.mark.skipif(have_numpy, reason="testing numpy ImportError")
    def test_numpy_import_warning(self):
        config.DS_numpy(True)
        config.use_IS_numpy = True
        with pytest.raises(ImportError):
            values.convert_DS_string(b"123.1", True)
        with pytest.raises(ImportError):
            values.convert_IS_string(b"123", True)


class TestDeferredRead:
    """Test that deferred data element reading (for large size)
    works as expected
    """

    # Copy one of test files and use temporarily, then later remove.
    def setup(self):
        self.testfile_name = ct_name + ".tmp"
        shutil.copyfile(ct_name, self.testfile_name)

    def teardown(self):
        if os.path.exists(self.testfile_name):
            os.remove(self.testfile_name)

    def test_time_check(self):
        """Deferred read warns if file has been modified"""
        ds = dcmread(self.testfile_name, defer_size="2 kB")
        from time import sleep

        sleep(0.1)
        with open(self.testfile_name, "r+") as f:
            f.write("\0")  # "touch" the file

        msg = r"Deferred read warning -- file modification time has changed"
        with pytest.warns(UserWarning, match=msg):
            ds.PixelData

    def test_file_exists(self):
        """Deferred read raises error if file no longer exists."""
        ds = dcmread(self.testfile_name, defer_size=2000)
        os.remove(self.testfile_name)
        with pytest.raises(IOError):
            ds.PixelData

    def test_values_identical(self):
        """Deferred values exactly matches normal read."""
        ds_norm = dcmread(self.testfile_name)
        ds_defer = dcmread(self.testfile_name, defer_size=2000)
        for data_elem in ds_norm:
            tag = data_elem.tag

            if have_numpy and isinstance(data_elem.value, numpy.ndarray):
                assert numpy.allclose(data_elem.value, ds_defer[tag].value)
            else:
                assert data_elem.value == ds_defer[tag].value

    def test_zipped_deferred(self):
        """Deferred values from a gzipped file works."""
        # Arose from issue 103 "Error for defer_size read of gzip file object"
        fobj = gzip.open(gzip_name)
        ds = dcmread(fobj, defer_size=1)
        fobj.close()
        # before the fix, this threw an error as file reading was not in
        # the right place, it was re-opened as a normal file, not a zip file
        ds.InstanceNumber

    def test_filelike_deferred(self):
        """Deferred values work with file-like objects."""
        with open(ct_name, "rb") as fp:
            data = fp.read()
        filelike = io.BytesIO(data)
        dataset = pydicom.dcmread(filelike, defer_size=1024)
        assert 32768 == len(dataset.PixelData)


class TestReadTruncatedFile:
    def testReadFileWithMissingPixelData(self):
        mr = dcmread(truncated_mr_name)
        mr.decode()
        assert "CompressedSamples^MR1" == mr.PatientName
        assert mr.PatientName == mr[0x10, 0x10].value
        DS = pydicom.valuerep.DS

        if have_numpy and config.use_DS_numpy:
            expected = numpy.array([0.3125, 0.3125])
            assert numpy.allclose(mr.PixelSpacing, expected)
        else:
            assert [DS("0.3125"), DS("0.3125")] == mr.PixelSpacing

    @pytest.mark.skipif(
        not have_numpy or have_gdcm_handler,
        reason="Missing numpy or GDCM present",
    )
    def testReadFileWithMissingPixelDataArray(self):
        mr = dcmread(truncated_mr_name)
        mr.decode()
        # Need to escape brackets
        msg = (
            r"The length of the pixel data in the dataset \(8130 bytes\) "
            r"doesn't match the expected length \(8192 bytes\). "
            r"The dataset may be corrupted or there may be an issue with "
            r"the pixel data handler."
        )
        with pytest.raises(ValueError, match=msg):
            mr.pixel_array


class TestFileLike:
    """Test that can read DICOM files with file-like object rather than
    filename
    """

    def test_read_file_given_file_object(self):
        """filereader: can read using already opened file............"""
        f = open(ct_name, "rb")
        ct = dcmread(f)
        # XXX Tests here simply repeat testCT -- perhaps should collapse
        # the code together?
        DS = pydicom.valuerep.DS

        got = ct.ImagePositionPatient
        if have_numpy and config.use_DS_numpy:
            expected = numpy.array([-158.135803, -179.035797, -75.699997])
            assert numpy.allclose(got, expected)
        else:
            expected = [DS("-158.135803"), DS("-179.035797"), DS("-75.699997")]
            assert expected == got

        assert "1.3.6.1.4.1.5962.2" == ct.file_meta.ImplementationClassUID
        value = ct.file_meta[0x2, 0x12].value
        assert ct.file_meta.ImplementationClassUID == value

        assert 128 == ct.Rows
        assert 128 == ct.Columns
        assert 16 == ct.BitsStored
        assert 128 * 128 * 2 == len(ct.PixelData)

        # Should also be able to close the file ourselves without
        # exception raised:
        f.close()

    def test_read_file_given_file_like_object(self):
        """filereader: can read using a file-like (BytesIO) file...."""
        with open(ct_name, "rb") as f:
            file_like = BytesIO(f.read())
        ct = dcmread(file_like)
        # Tests here simply repeat some of testCT test
        got = ct.ImagePositionPatient
        DS = pydicom.valuerep.DS

        if have_numpy and config.use_DS_numpy:
            expected = numpy.array([-158.135803, -179.035797, -75.699997])
            assert numpy.allclose(got, expected)
        else:
            expected = [DS("-158.135803"), DS("-179.035797"), DS("-75.699997")]
            assert expected == got

        assert 128 * 128 * 2 == len(ct.PixelData)

        # Should also be able to close the file ourselves without
        # exception raised:
        file_like.close()


class TestDataElementGenerator:
    """Test filereader.data_element_generator"""

    def test_little_endian_explicit(self):
        """Test reading little endian explicit VR data"""
        # (0010, 0010) PatientName PN 6 ABCDEF
        bytestream = b"\x10\x00\x10\x00" b"PN" b"\x06\x00" b"ABCDEF"
        fp = BytesIO(bytestream)
        # fp, is_implicit_VR, is_little_endian,
        gen = data_element_generator(fp, False, True)
        elem = DataElement(0x00100010, "PN", "ABCDEF")
        assert elem == DataElement_from_raw(next(gen), "ISO_IR 100")

    def test_little_endian_implicit(self):
        """Test reading little endian implicit VR data"""
        # (0010, 0010) PatientName PN 6 ABCDEF
        bytestream = b"\x10\x00\x10\x00" b"\x06\x00\x00\x00" b"ABCDEF"
        fp = BytesIO(bytestream)
        gen = data_element_generator(
            fp, is_implicit_VR=True, is_little_endian=True
        )
        elem = DataElement(0x00100010, "PN", "ABCDEF")
        assert elem == DataElement_from_raw(next(gen), "ISO_IR 100")

    def test_big_endian_explicit(self):
        """Test reading big endian explicit VR data"""
        # (0010, 0010) PatientName PN 6 ABCDEF
        bytestream = b"\x00\x10\x00\x10" b"PN" b"\x00\x06" b"ABCDEF"
        fp = BytesIO(bytestream)
        # fp, is_implicit_VR, is_little_endian,
        gen = data_element_generator(fp, False, False)
        elem = DataElement(0x00100010, "PN", "ABCDEF")
        assert elem == DataElement_from_raw(next(gen), "ISO_IR 100")


def test_read_dicomdir_deprecated():
    """Test deprecation warning for read_dicomdir()."""
    msg = (
        r"'read_dicomdir\(\)' is deprecated and will be removed in v3.0, use "
        r"'dcmread\(\)' instead"
    )
    with pytest.warns(DeprecationWarning, match=msg):
        ds = read_dicomdir(get_testdata_file("DICOMDIR"))


def test_read_file_deprecated():
    """Test deprecation warning for read_file()."""
    if sys.version_info[:2] < (3, 7):
        from pydicom.filereader import read_file
    else:
        msg = (
            r"'read_file' is deprecated and will be removed in v3.0, use "
            r"'dcmread' instead"
        )
        with pytest.warns(DeprecationWarning, match=msg):
            from pydicom.filereader import read_file

    assert read_file == dcmread



# Copyright 2020 pydicom authors. See LICENSE file for details.
"""Tests for command-line interface"""

from argparse import ArgumentTypeError

import pytest

from pydicom.cli.main import (
    filespec_parser, eval_element, main, filespec_parts
)


bad_elem_specs = (
    "extra:colon",
    "no_callable()",
    "no_equals = ",
    "BeamSequence[0]extra",  # must match to end of string
    "BeamSequence[x]",  # index must be an int
)

missing_elements = (
    "NotThere",
    "BeamSequenceXX",
    "BeamDose",  # valid keyword but not at top level
)

bad_indexes = (
    "BeamSequence[42]",
    "BeamSequence[-42]",
)


class TestFilespec:
    @pytest.mark.parametrize("bad_spec", bad_elem_specs)
    def test_syntax(self, bad_spec):
        """Invalid syntax for for CLI file:element spec raises error"""
        with pytest.raises(ArgumentTypeError, match=r".* syntax .*"):
            filespec_parser(f"pydicom::rtplan.dcm::{bad_spec}")

    @pytest.mark.parametrize("missing_element", missing_elements)
    def test_elem_not_exists(self, missing_element):
        """CLI filespec elements not in the dataset raise an error"""
        with pytest.raises(
            ArgumentTypeError, match=r".* is not in the dataset"
        ):
            filespec_parser(f"pydicom::rtplan.dcm::{missing_element}")

    @pytest.mark.parametrize("bad_index", bad_indexes)
    def test_bad_index(self, bad_index):
        """CLI filespec elements with an invalid index raise an error"""
        with pytest.raises(ArgumentTypeError, match=r".* index error"):
            filespec_parser(f"pydicom::rtplan.dcm::{bad_index}")

    def test_offers_pydicom_testfile(self):
        """CLI message offers pydicom data file if file not found"""
        with pytest.raises(
            ArgumentTypeError, match=r".*pydicom::rtplan\.dcm.*is available.*"
        ):
            filespec_parser(f"rtplan.dcm")

    def test_colons(self):
        """CLI filespec with a colon in filename works correctly"""
        expected = ("", r"c:\test.dcm", "")
        assert expected == filespec_parts(r"c:\test.dcm")

        expected = ("pydicom", r"c:\test.dcm", "")
        assert expected == filespec_parts(r"pydicom::c:\test.dcm")

        filespec = r"pydicom::c:\test.dcm::StudyDate"
        expected = ("pydicom", r"c:\test.dcm", "StudyDate")
        assert expected == filespec_parts(filespec)

        filespec = r"c:\test.dcm::StudyDate"
        expected = ("", r"c:\test.dcm", "StudyDate")
        assert expected == filespec_parts(filespec)


class TestFilespecElementEval:
    # Load plan once
    plan, _ = filespec_parser("pydicom::rtplan.dcm")[0]

    def test_correct_values(self):
        """CLI produces correct evaluation of requested element"""
        # A nested data element
        elem_str = "BeamSequence[0].ControlPointSequence[0].NominalBeamEnergy"
        elem_val = eval_element(self.plan, elem_str)
        assert 6.0 == elem_val

        # A nested Sequence item
        elem_str = "BeamSequence[0].ControlPointSequence[0]"
        elem_val = eval_element(self.plan, elem_str)
        assert 6.0 == elem_val.NominalBeamEnergy

        # A nested Sequence itself
        elem_str = "BeamSequence[0].ControlPointSequence"
        elem_val = eval_element(self.plan, elem_str)
        assert 6.0 == elem_val[0].NominalBeamEnergy

        # A non-nested data element
        elem_str = "PatientID"
        elem_val = eval_element(self.plan, elem_str)
        assert "id00001" == elem_val

        # The file_meta or file_meta data element
        elem_str = "file_meta"
        elem_val = eval_element(self.plan, elem_str)
        assert "RT Plan Storage" == elem_val.MediaStorageSOPClassUID.name

        elem_str = "file_meta.MediaStorageSOPClassUID"
        elem_val = eval_element(self.plan, elem_str)
        assert "RT Plan Storage" == elem_val.name


class TestCLIcall:
    """Test calls to `pydicom` command-line interface"""

    def test_bare_command(self, capsys):
        """CLI `pydicom` with no arguments displays help"""
        main([])
        out, _ = capsys.readouterr()
        assert out.startswith("usage: pydicom [-h] {")

    def test_codify_command(self, capsys):
        """CLI `codify` command prints correct output"""

        # With private elements
        main("codify -p pydicom::nested_priv_SQ.dcm".split())
        out, _ = capsys.readouterr()
        assert "add_new((0x0001, 0x0001)" in out

        # Without private elements
        main("codify pydicom::nested_priv_SQ.dcm".split())
        out, _ = capsys.readouterr()
        assert "add_new((0x0001, 0x0001)" not in out

    def test_codify_data_element(self, capsys):
        """CLI `codify` command raises error if not a Dataset"""
        with pytest.raises(NotImplementedError):
            main("codify pydicom::rtplan.dcm::RTPlanLabel".split())

    def test_help(self, capsys):
        """CLI `help` command gives expected output"""
        # With subcommand
        main("help show".split())
        out, err = capsys.readouterr()
        assert out.startswith("usage: pydicom show [-h] [")
        assert err == ""

        # No subcommand following
        main(["help"])
        out, _ = capsys.readouterr()
        assert "Available subcommands:" in out

        # Non-existent subcommand following
        main("help DoesntExist".split())
        out, _ = capsys.readouterr()
        assert "Available subcommands:" in out

    def test_show_command(self, capsys):
        """CLI `show` command prints correct output"""
        main("show pydicom::MR_small_RLE.dcm".split())
        out, err = capsys.readouterr()

        assert "Instance Creation Date              DA: '20040826'" in out
        assert out.endswith("OB: Array of 126 elements\n")
        assert err == ""

        # Get a specific data element
        main("show pydicom::MR_small_RLE.dcm::LargestImagePixelValue".split())
        out, _ = capsys.readouterr()
        assert "4000" == out.strip()

    def test_show_options(self, capsys):
        """CLI `show` command with options prints correct output"""
        # Quiet option, image file
        main("show -q pydicom::MR_small_RLE.dcm".split())
        out, err = capsys.readouterr()

        assert out.startswith("SOPClassUID: MR Image Storage")
        assert out.endswith("Rows: 64\nColumns: 64\nSliceLocation: 0.0000\n")
        assert err == ""

        # 'Quiet' option, RTPLAN file
        main("show -q pydicom::rtplan.dcm".split())
        out, err = capsys.readouterr()
        assert out.endswith(
            "Beam 1 'Field 1' TREATMENT STATIC PHOTON energy 6.00000000000000 "
            "gantry 0.0, coll 0.0, couch 0.0 "
            "(0 wedges, 0 comps, 0 boli, 0 blocks)\n"
        )
        assert err == ""

        # Top-level-only option, also different file for more variety
        main("show -t pydicom::nested_priv_SQ.dcm".split())
        out, err = capsys.readouterr()
        assert "(0001, 0001)  Private Creator" in out
        assert "UN: b'Nested SQ'" not in out
        assert err == ""

        # Exclude private option
        main("show -x pydicom::nested_priv_SQ.dcm".split())
        out, err = capsys.readouterr()
        assert "(0001, 0001)  Private Creator" not in out
        assert err == ""



# Copyright 2008-2018 pydicom authors. See LICENSE file for details.
"""Unit tests for the pydicom.tag module."""

import pytest

from pydicom.tag import BaseTag, Tag, TupleTag, tag_in_exception


class TestBaseTag:
    """Test the BaseTag class."""
    def test_le_same_class(self):
        """Test __le__ of two classes with same type."""
        assert BaseTag(0x00000000) <= BaseTag(0x00000001)
        assert BaseTag(0x00000001) <= BaseTag(0x00000001)
        assert not BaseTag(0x00000001) <= BaseTag(0x00000000)

    def test_le_diff_class(self):
        """Test __le__ of two classes with different type."""
        assert BaseTag(0x00000000) <= 1
        assert BaseTag(0x00000001) <= 1
        assert not BaseTag(0x00000001) <= 0

    def test_le_subclass(self):
        """Test __le__ of two classes with one as a subclass."""
        class BaseTagPlus(BaseTag):
            pass
        assert BaseTagPlus(0x00000000) <= BaseTag(0x00000001)
        assert BaseTagPlus(0x00000001) <= BaseTag(0x00000001)
        assert not BaseTagPlus(0x00000001) <= BaseTag(0x00000000)

    def test_le_tuple(self):
        """Test __le__ of tuple with BaseTag."""
        assert BaseTag(0x00010001) <= (0x0001, 0x0002)
        assert BaseTag(0x00010002) <= (0x0001, 0x0002)
        assert not BaseTag(0x00010002) <= (0x0001, 0x0001)

    def test_le_raises(self):
        """Test __le__ raises TypeError when comparing to non numeric."""
        def test_raise():
            BaseTag(0x00010002) <= 'Somethin'
        pytest.raises(TypeError, test_raise)

    def test_lt_same_class(self):
        """Test __lt__ of two classes with same type."""
        assert BaseTag(0x00000000) < BaseTag(0x00000001)
        assert not BaseTag(0x00000001) < BaseTag(0x00000001)
        assert not BaseTag(0x00000001) < BaseTag(0x00000000)

    def test_lt_diff_class(self):
        """Test __lt__ of two classes with different type."""
        assert BaseTag(0x00000000) < 1
        assert not BaseTag(0x00000001) < 1
        assert not BaseTag(0x00000001) < 0

    def test_lt_subclass(self):
        """Test __lt__ of two classes with one as a subclass."""
        class BaseTagPlus(BaseTag):
            pass
        assert BaseTagPlus(0x00000000) < BaseTag(0x00000001)
        assert not BaseTagPlus(0x00000001) < BaseTag(0x00000001)
        assert not BaseTagPlus(0x00000001) < BaseTag(0x00000000)

    def test_lt_tuple(self):
        """Test __lt__ of tuple with BaseTag."""
        assert BaseTag(0x00010001) < (0x0001, 0x0002)
        assert not BaseTag(0x00010002) < (0x0001, 0x0002)
        assert not BaseTag(0x00010002) < (0x0001, 0x0001)

    def test_lt_raises(self):
        """Test __lt__ raises TypeError when comparing to non numeric."""
        def test_raise():
            BaseTag(0x00010002) < 'Somethin'
        pytest.raises(TypeError, test_raise)

    def test_ge_same_class(self):
        """Test __ge__ of two classes with same type."""
        assert not BaseTag(0x00000000) >= BaseTag(0x00000001)
        assert BaseTag(0x00000001) >= BaseTag(0x00000001)
        assert BaseTag(0x00000001) >= BaseTag(0x00000000)

    def test_ge_diff_class(self):
        """Test __ge__ of two classes with different type."""
        assert not BaseTag(0x00000000) >= 1
        assert BaseTag(0x00000001) >= 1
        assert BaseTag(0x00000001) >= 0

    def test_ge_subclass(self):
        """Test __ge__ of two classes with one as a subclass."""
        class BaseTagPlus(BaseTag):
            pass
        assert not BaseTagPlus(0x00000000) >= BaseTag(0x00000001)
        assert BaseTagPlus(0x00000001) >= BaseTag(0x00000001)
        assert BaseTagPlus(0x00000001) >= BaseTag(0x00000000)

    def test_ge_tuple(self):
        """Test __ge__ of tuple with BaseTag."""
        assert not BaseTag(0x00010001) >= (0x0001, 0x0002)
        assert BaseTag(0x00010002) >= (0x0001, 0x0002)
        assert BaseTag(0x00010002) >= (0x0001, 0x0001)

    def test_ge_raises(self):
        """Test __ge__ raises TypeError when comparing to non numeric."""
        def test_raise():
            BaseTag(0x00010002) >= 'AGHIJJJJ'
        pytest.raises(TypeError, test_raise)

    def test_gt_same_class(self):
        """Test __gt__ of two classes with same type."""
        assert not BaseTag(0x00000000) > BaseTag(0x00000001)
        assert not BaseTag(0x00000001) > BaseTag(0x00000001)
        assert BaseTag(0x00000001) > BaseTag(0x00000000)

    def test_gt_diff_class(self):
        """Test __gt__ of two classes with different type."""
        assert not BaseTag(0x00000000) > 1
        assert not BaseTag(0x00000001) > 1
        assert BaseTag(0x00000001) > 0

    def test_gt_subclass(self):
        """Test __gt__ of two classes with one as a subclass."""
        class BaseTagPlus(BaseTag):
            pass
        assert not BaseTagPlus(0x00000000) > BaseTag(0x00000001)
        assert not BaseTagPlus(0x00000001) > BaseTag(0x00000001)
        assert BaseTagPlus(0x00000001) > BaseTag(0x00000000)

    def test_gt_tuple(self):
        """Test __gt__ of tuple with BaseTag."""
        assert not BaseTag(0x00010001) > (0x0001, 0x0002)
        assert not BaseTag(0x00010002) > (0x0001, 0x0002)
        assert BaseTag(0x00010002) > (0x0001, 0x0001)

    def test_gt_raises(self):
        """Test __gt__ raises TypeError when comparing to non numeric."""
        def test_raise():
            BaseTag(0x00010002) > 'BLUH'
        pytest.raises(TypeError, test_raise)

    def test_eq_same_class(self):
        """Test __eq__ of two classes with same type."""
        assert BaseTag(0x00000000) == BaseTag(0x00000000)
        assert not BaseTag(0x00000001) == BaseTag(0x00000000)

    def test_eq_diff_class(self):
        """Test __eq__ of two classes with different type."""
        # Make sure to test BaseTag.__eq__() not int.__eq__()
        assert BaseTag(0x00000000) == 0
        assert not BaseTag(0x00000001) == 0

    def test_eq_subclass(self):
        """Test __eq__ of two classes with one as a subclass."""
        class BaseTagPlus(BaseTag):
            pass
        assert BaseTagPlus(0x00000000) == BaseTag(0x00000000)
        assert not BaseTagPlus(0x00000001) == BaseTag(0x00000000)

    def test_eq_tuple(self):
        """Test __eq__ of tuple with BaseTag."""
        # Make sure to test BaseTag.__eq__() not tuple.__eq__()
        assert BaseTag(0x00010002) == (0x0001, 0x0002)
        assert not BaseTag(0x00010001) == (0x0001, 0x0002)

    def test_eq_non_numeric(self):
        """Test __eq__ of non numeric with BaseTag."""
        assert not BaseTag(0x00010002) == 'eraa'

    def test_ne_same_class(self):
        """Test __ne__ of two classes with same type."""
        assert not BaseTag(0x00000000) != BaseTag(0x00000000)
        assert BaseTag(0x00000001) != BaseTag(0x00000000)

    def test_ne_diff_class(self):
        """Test __ne__ of two classes with different type."""
        # Make sure to test BaseTag.__ne__() not int.__ne__()
        assert not BaseTag(0x00000000) != 0
        assert BaseTag(0x00000001) != 0

    def test_ne_subclass(self):
        """Test __ne__ of two classes with one as a subclass."""
        class BaseTagPlus(BaseTag):
            pass
        assert not BaseTagPlus(0x00000000) != BaseTag(0x00000000)
        assert BaseTagPlus(0x00000001) != BaseTag(0x00000000)

    def test_ne_tuple(self):
        """Test __ne__ of tuple with BaseTag."""
        # Make sure to test BaseTag.__ne__() not tuple.__ne__()
        assert not BaseTag(0x00010002) != (0x0001, 0x0002)
        assert BaseTag(0x00010001) != (0x0001, 0x0002)

    def test_ne_non_numeric(self):
        """Test __ne__ of non numeric with BaseTag."""
        assert BaseTag(0x00010002) != 'aaag'

    def test_hash(self):
        """Test hash of BaseTag class."""
        assert hash(BaseTag(0x00010001)) == hash(BaseTag(0x00010001))
        assert hash(BaseTag(0x00010001)) != hash(BaseTag(0x00010002))
        assert hash(BaseTag(0x00020001)) != hash(BaseTag(0x00010002))

    def test_str(self):
        """Test str(BaseTag) produces correct value."""
        assert '(0000, 0000)' == str(BaseTag(0x00000000))
        assert '(0001, 0002)' == str(BaseTag(0x00010002))
        assert '(1000, 2000)' == str(BaseTag(0x10002000))
        assert '(ffff, fffe)' == str(BaseTag(0xFFFFFFFE))

    def test_group(self):
        """Test BaseTag.group returns correct values."""
        assert 0x0000 == BaseTag(0x00000001).group
        assert 0x0002 == BaseTag(0x00020001).group
        assert 0xFFFF == BaseTag(0xFFFF0001).group

    def test_element(self):
        """Test BaseTag.element returns correct values."""
        assert 0x0000 == BaseTag(0x00010000).element
        assert 0x0002 == BaseTag(0x00010002).element
        assert 0xFFFF == BaseTag(0x0001FFFF).element

    def test_private(self):
        """Test BaseTag.is_private returns correct values."""
        # Odd groups private
        assert BaseTag(0x00010001).is_private
        # Even groups not private
        assert not BaseTag(0x00020001).is_private
        # Group 0 not private
        assert not BaseTag(0x00000001).is_private

    def test_private_creator(self):
        """Test BaseTag.is_private_creator returns correct values."""
        # Non-private tag
        assert not BaseTag(0x00080010).is_private_creator
        # private creator have element 0x0010 - 0x00FF
        assert not BaseTag(0x0009000F).is_private_creator
        assert BaseTag(0x00090010).is_private_creator
        assert BaseTag(0x000900FF).is_private_creator
        assert not BaseTag(0x00090100).is_private_creator


class TestTag:
    """Test the Tag method."""
    def test_tag_single_int(self):
        """Test creating a Tag from a single int."""
        assert Tag(0x0000) == BaseTag(0x00000000)
        assert Tag(10) == BaseTag(0x0000000A)
        assert Tag(0xFFFF) == BaseTag(0x0000FFFF)
        assert Tag(0x00010002) == BaseTag(0x00010002)

        # Must be 32-bit
        pytest.raises(OverflowError, Tag, 0xFFFFFFFF1)
        # Must be positive
        pytest.raises(ValueError, Tag, -1)

    def test_tag_single_tuple(self):
        """Test creating a Tag from a single tuple."""
        assert Tag((0x0000, 0x0000)) == BaseTag(0x00000000)
        assert Tag((0x22, 0xFF)) == BaseTag(0x002200FF)
        assert Tag((14, 0xF)) == BaseTag(0x000E000F)
        assert Tag((0x1000, 0x2000)) == BaseTag(0x10002000)
        assert Tag(('0x01', '0x02')) == BaseTag(0x00010002)

        # Must be 2 tuple
        pytest.raises(ValueError, Tag, (0x1000, 0x2000, 0x0030))
        pytest.raises(ValueError, Tag, ('0x10', '0x20', '0x03'))
        # Must be 32-bit
        pytest.raises(OverflowError, Tag, (0xFFFF, 0xFFFF1))
        pytest.raises(OverflowError, Tag, ('0xFFFF', '0xFFFF1'))
        # Must be positive
        pytest.raises(ValueError, Tag, (-1, 0))
        pytest.raises(ValueError, Tag, (0, -1))
        pytest.raises(ValueError, Tag, ('0x0', '-0x1'))
        pytest.raises(ValueError, Tag, ('-0x1', '0x0'))
        # Can't have second parameter
        msg = r"Unable to create an element tag from '\(\(1, 2\), 1\)'"
        with pytest.raises(TypeError, match=msg):
            Tag((0x01, 0x02), 0x01)
        pytest.raises(TypeError, Tag, (0x01, 0x02), '0x01')
        pytest.raises(TypeError, Tag, ('0x01', '0x02'), '0x01')
        pytest.raises(TypeError, Tag, ('0x01', '0x02'), 0x01)

    def test_tag_single_list(self):
        """Test creating a Tag from a single list."""
        assert Tag([0x0000, 0x0000]) == BaseTag(0x00000000)
        assert Tag([0x99, 0xFE]) == BaseTag(0x009900FE)
        assert Tag([15, 0xE]) == BaseTag(0x000F000E)
        assert Tag([0x1000, 0x2000]) == BaseTag(0x10002000)
        assert Tag(['0x01', '0x02']) == BaseTag(0x00010002)

        # Must be 2 list
        pytest.raises(ValueError, Tag, [0x1000, 0x2000, 0x0030])
        pytest.raises(ValueError, Tag, ['0x10', '0x20', '0x03'])
        pytest.raises(ValueError, Tag, [0x1000])
        pytest.raises(ValueError, Tag, ['0x10'])

        # Must be int or string
        msg = (
            r"Unable to create an element tag from '\[1.0, 2.0\]': both "
            r"arguments must be the same type and str or int"
        )
        with pytest.raises(TypeError, match=msg):
            Tag([1., 2.])

        # Must be 32-bit
        pytest.raises(OverflowError, Tag, [65536, 0])
        pytest.raises(OverflowError, Tag, [0, 65536])
        pytest.raises(OverflowError, Tag, ('0xFFFF', '0xFFFF1'))
        # Must be positive
        pytest.raises(ValueError, Tag, [-1, 0])
        pytest.raises(ValueError, Tag, [0, -1])
        pytest.raises(ValueError, Tag, ('0x0', '-0x1'))
        pytest.raises(ValueError, Tag, ('-0x1', '0x0'))
        # Can't have second parameter
        msg = r"Unable to create an element tag from '\(\[1, 2\], 1\)'"
        with pytest.raises(TypeError, match=msg):
            Tag([0x01, 0x02], 0x01)

        pytest.raises(TypeError, Tag, [0x01, 0x02], '0x01')
        pytest.raises(TypeError, Tag, ['0x01', '0x02'], '0x01')
        pytest.raises(TypeError, Tag, ['0x01', '0x02'], 0x01)

    def test_tag_single_str(self):
        """Test creating a Tag from a single str."""
        assert Tag('0x10002000') == BaseTag(0x10002000)
        assert Tag('0x2000') == BaseTag(0x00002000)
        assert Tag('15') == BaseTag(0x00000015)
        assert Tag('0xF') == BaseTag(0x0000000F)
        assert Tag("PatientName") == BaseTag(0x00100010)

        # Must be 32-bit
        pytest.raises(OverflowError, Tag, '0xFFFFFFFF1')
        # Must be positive
        pytest.raises(ValueError, Tag, '-0x01')
        # Must be numeric str or DICOM keyword
        pytest.raises(ValueError, Tag, 'hello')

    def test_tag_double_str(self):
        """Test creating a Tag from two str."""
        assert Tag('0x1000', '0x2000') == BaseTag(0x10002000)
        assert Tag('0x10', '0x20') == BaseTag(0x00100020)
        assert Tag('15', '0') == BaseTag(0x00150000)
        assert Tag('0xF', '0') == BaseTag(0x000F0000)

        # Must be 32-bit
        pytest.raises(OverflowError, Tag, '0xFFFF1', '0')
        pytest.raises(OverflowError, Tag, '0', '0xFFFF1')
        # Must be positive
        pytest.raises(ValueError, Tag, '-0x01', '0')
        pytest.raises(ValueError, Tag, '0', '-0x01')
        pytest.raises(ValueError, Tag, '-1', '-0x01')
        # Must both be str
        pytest.raises(TypeError, Tag, '0x01', 0)
        pytest.raises(TypeError, Tag, 0, '0x01')

    def test_tag_double_int(self):
        """Test creating a Tag from a two ints."""
        assert Tag(0x0000, 0x0000) == BaseTag(0x00000000)
        assert Tag(2, 0) == BaseTag(0x00020000)
        assert Tag(2, 0).elem == 0x0000
        assert Tag(0x99, 0xFE) == BaseTag(0x009900FE)
        assert Tag(15, 14) == BaseTag(0x000F000E)
        assert Tag(0x1000, 0x2000) == BaseTag(0x10002000)

        # Must be 32-bit
        pytest.raises(OverflowError, Tag, 65536, 0)
        pytest.raises(OverflowError, Tag, 0, 65536)
        # Must be positive
        pytest.raises(ValueError, Tag, -1, 0)
        pytest.raises(ValueError, Tag, 0, -1)
        pytest.raises(ValueError, Tag, -65535, -1)


class TestTupleTag:
    """Test the TupleTag method."""
    def test_tuple_tag(self):
        """Test quick tag construction with TupleTag."""
        assert TupleTag((0xFFFF, 0xFFee)) == BaseTag(0xFFFFFFEE)


class TestTagInException:
    """Test the tag_in_exception method."""
    def test_raise_exception(self):
        """"""
        def test():
            tag = Tag(0x00100010)
            with tag_in_exception(tag) as tag:
                raise ValueError('Test message.')
        pytest.raises(ValueError, test)
